"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.ssh_auth as ssh_auth
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {ssh_auth: {}}


def test_present():
    """
    Test to verifies that the specified SSH key
    is present for the specified user.
    """
    name = "sshkeys"
    user = "root"
    source = "salt://ssh_keys/id_rsa.pub"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(return_value="exists")
    mock_data = MagicMock(side_effect=["replace", "new"])
    with patch.dict(
        ssh_auth.__salt__, {"ssh.check_key": mock, "ssh.set_auth_key": mock_data}
    ):
        with patch.dict(ssh_auth.__opts__, {"test": True}):
            comt = "The authorized host key sshkeys is already present for user root"
            ret.update({"comment": comt})
            assert ssh_auth.present(name, user, source) == ret

        with patch.dict(ssh_auth.__opts__, {"test": False}):
            comt = "The authorized host key sshkeys for user root was updated"
            ret.update({"comment": comt, "changes": {name: "Updated"}})
            assert ssh_auth.present(name, user, source) == ret

            comt = "The authorized host key sshkeys for user root was added"
            ret.update({"comment": comt, "changes": {name: "New"}})
            assert ssh_auth.present(name, user, source) == ret


def test_absent():
    """
    Test to verifies that the specified SSH key is absent.
    """
    name = "sshkeys"
    user = "root"
    source = "salt://ssh_keys/id_rsa.pub"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    mock = MagicMock(
        side_effect=["User authorized keys file not present", "Key removed"]
    )
    mock_up = MagicMock(side_effect=["update", "updated"])
    with patch.dict(
        ssh_auth.__salt__, {"ssh.rm_auth_key": mock, "ssh.check_key": mock_up}
    ):
        with patch.dict(ssh_auth.__opts__, {"test": True}):
            comt = "Key sshkeys for user root is set for removal"
            ret.update({"comment": comt})
            assert ssh_auth.absent(name, user, source) == ret

            comt = "Key is already absent"
            ret.update({"comment": comt, "result": True})
            assert ssh_auth.absent(name, user, source) == ret

        with patch.dict(ssh_auth.__opts__, {"test": False}):
            comt = "User authorized keys file not present"
            ret.update({"comment": comt, "result": False})
            assert ssh_auth.absent(name, user, source) == ret

            comt = "Key removed"
            ret.update({"comment": comt, "result": True, "changes": {name: "Removed"}})
            assert ssh_auth.absent(name, user, source) == ret


def test_manage():
    """
    Test to verifies that the specified SSH key is absent.
    """
    user = "root"
    ret = {"name": "", "changes": {}, "result": None, "comment": ""}

    mock_rm = MagicMock(
        side_effect=["User authorized keys file not present", "Key removed"]
    )
    mock_up = MagicMock(side_effect=["update", "updated"])
    mock_set = MagicMock(side_effect=["replace", "new"])
    mock_keys = MagicMock(
        return_value={
            "somekey": {
                "enc": "ssh-rsa",
                "comment": "user@host",
                "options": [],
                "fingerprint": "b7",
            }
        }
    )
    with patch.dict(
        ssh_auth.__salt__,
        {
            "ssh.rm_auth_key": mock_rm,
            "ssh.set_auth_key": mock_set,
            "ssh.check_key": mock_up,
            "ssh.auth_keys": mock_keys,
        },
    ):
        with patch("salt.states.ssh_auth.present") as call_mocked_present:
            mock_present = {"comment": "", "changes": {}, "result": None}
            call_mocked_present.return_value = mock_present
            with patch.dict(ssh_auth.__opts__, {"test": True}):
                assert ssh_auth.manage("sshid", ["somekey"], user) == ret

                comt = "somekey Key set for removal"
                ret.update({"comment": comt})
                assert ssh_auth.manage("sshid", [], user) == ret

        with patch("salt.states.ssh_auth.present") as call_mocked_present:
            mock_present = {"comment": "", "changes": {}, "result": True}
            call_mocked_present.return_value = mock_present
            with patch.dict(ssh_auth.__opts__, {"test": False}):
                ret = {"name": "", "changes": {}, "result": True, "comment": ""}
                assert ssh_auth.manage("sshid", ["somekey"], user) == ret

                with patch("salt.states.ssh_auth.absent") as call_mocked_absent:
                    mock_absent = {"comment": "Key removed"}
                    call_mocked_absent.return_value = mock_absent
                    ret.update(
                        {
                            "comment": "",
                            "result": True,
                            "changes": {"somekey": "Key removed"},
                        }
                    )
                    assert ssh_auth.manage("sshid", ["addkey"], user) == ret

        # add a key
        with patch("salt.states.ssh_auth.present") as call_mocked_present:
            mock_present = {
                "comment": (
                    "The authorized host key newkey for user {} was added".format(user)
                ),
                "changes": {"newkey": "New"},
                "result": True,
            }
            call_mocked_present.return_value = mock_present
            with patch.dict(ssh_auth.__opts__, {"test": False}):
                ret = {
                    "name": "",
                    "changes": {"newkey": "New"},
                    "result": True,
                    "comment": "",
                }
                assert ssh_auth.manage("sshid", ["newkey", "somekey"], user) == ret



import logging

import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def data_module(modules):
    return modules.data


def test_clear(data_module):
    """
    data.clear
    """
    ret = data_module.update("foo", "bar")
    assert ret

    ret = data_module.clear()
    assert ret

    ret = data_module.items()
    assert ret == {}


def test_dump(data_module):
    """
    data.dump
    """
    ret = data_module.update("foo", "bar")
    assert ret

    ret = data_module.items()
    assert ret == {"foo": "bar"}

    ret = data_module.dump("{'bar': 'baz'}")
    assert ret

    ret = data_module.items()
    assert ret == {"bar": "baz"}


def test_cas(data_module):
    """
    data.cas
    """
    ret = data_module.update("foo", "bar")
    assert ret

    ret = data_module.items()
    assert ret == {"foo": "bar"}

    ret = data_module.cas("foo", "baz", "bar")
    assert ret

    ret = data_module.items()
    assert ret == {"foo": "baz"}


def test_pop(data_module):
    """
    data.pop
    """
    ret = data_module.update("foo", "bar")
    assert ret

    ret = data_module.update("bar", "baz")
    assert ret

    ret = data_module.items()
    assert ret == {"foo": "bar", "bar": "baz"}

    ret = data_module.pop("bar")
    assert ret == "baz"

    ret = data_module.items()
    assert ret == {"foo": "bar"}


def test_get(data_module):
    """
    data.get
    """
    ret = data_module.update("foo", "bar")
    assert ret

    ret = data_module.update("bar", "baz")
    assert ret

    ret = data_module.get("foo")
    assert ret == "bar"

    ret = data_module.get(["foo", "bar"])
    assert ret == ["bar", "baz"]


def test_items(data_module):
    """
    data.items
    """
    ret = data_module.update("foo", "bar")
    assert ret

    ret = data_module.items()
    assert ret == {"foo": "bar"}


def test_values(data_module):
    """
    data.values
    """
    ret = data_module.update("foo", "bar")
    assert ret

    ret = data_module.values()
    assert ret == ["bar"]


def test_keys(data_module):
    """
    data.keys
    """
    ret = data_module.update("foo", "bar")
    assert ret

    ret = data_module.keys()
    assert ret == ["foo"]


def test_has_key(data_module):
    """
    data.has_key
    """
    ret = data_module.update("foo", "bar")
    assert ret

    ret = data_module.has_key("foo")  # pylint: disable=.has_key-is-deprecated-use-in
    assert ret

    ret = data_module.has_key("bar")  # pylint: disable=.has_key-is-deprecated-use-in

    assert not ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.boto_asg as boto_asg
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {boto_asg: {}}


def test_present():
    """
    Test to ensure the autoscale group exists.
    """
    name = "myasg"
    launch_config_name = "mylc"
    availability_zones = ["us-east-1a", "us-east-1b"]
    min_size = 1
    max_size = 1

    ret = {"name": name, "result": None, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[False, {"min_size": 2}, [""]])
    with patch.dict(boto_asg.__salt__, {"boto_asg.get_config": mock}):
        with patch.dict(boto_asg.__opts__, {"test": True}):
            comt = "Autoscale group set to be created."
            ret.update({"comment": comt})
            with patch.dict(
                boto_asg.__salt__, {"config.option": MagicMock(return_value={})}
            ):
                assert (
                    boto_asg.present(
                        name,
                        launch_config_name,
                        availability_zones,
                        min_size,
                        max_size,
                    )
                    == ret
                )

            def magic_side_effect(value):
                if isinstance(value, int):
                    if value == 1:
                        return 4
                    return value
                return ""

            comt = "Autoscale group set to be updated."
            ret.update({"comment": comt, "result": None})
            ret.update({"changes": {"new": {"min_size": 4}, "old": {"min_size": 2}}})
            utils_ordered_mock = MagicMock(side_effect=magic_side_effect)
            with patch.dict(
                boto_asg.__salt__, {"config.option": MagicMock(return_value={})}
            ):
                with patch.dict(
                    boto_asg.__utils__, {"boto3.ordered": utils_ordered_mock}
                ):
                    call_ret = boto_asg.present(
                        name,
                        launch_config_name,
                        availability_zones,
                        min_size,
                        max_size,
                    )
                    assert call_ret == ret

            with patch.dict(
                boto_asg.__salt__, {"config.option": MagicMock(return_value={})}
            ):
                with patch.dict(
                    boto_asg.__utils__,
                    {"boto3.ordered": MagicMock(return_value="")},
                ):
                    comt = "Autoscale group present. "
                    ret.update({"comment": comt, "result": True})
                    ret.update({"changes": {}})
                    assert (
                        boto_asg.present(
                            name,
                            launch_config_name,
                            availability_zones,
                            min_size,
                            max_size,
                        )
                        == ret
                    )


def test_absent():
    """
    Test to ensure the named autoscale group is deleted.
    """
    name = "myasg"

    ret = {"name": name, "result": None, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[True, False])
    with patch.dict(boto_asg.__salt__, {"boto_asg.get_config": mock}):
        with patch.dict(boto_asg.__opts__, {"test": True}):
            comt = "Autoscale group set to be deleted."
            ret.update({"comment": comt})
            assert boto_asg.absent(name) == ret

            comt = "Autoscale group does not exist."
            ret.update({"comment": comt, "result": True})
            assert boto_asg.absent(name) == ret



"""
Monitor Web Server with Uptime
==============================

`Uptime <https://github.com/fzaninotto/uptime>`_ is an open source
remote monitoring application using Node.js, MongoDB, and Twitter
Bootstrap.

.. warning::

    This state module is beta. It might be changed later to include
    more or less automation.

.. note::

    This state module requires a pillar to specify the location of
    your uptime install

    .. code-block:: yaml

        uptime:
          application_url: "http://uptime-url.example.org"

Example:

.. code-block:: yaml

    url:
      uptime.monitored
    url/sitemap.xml:
      uptime.monitored:
         - polling: 600 # every hour

"""


def __virtual__():
    """
    Only load if the uptime module is present
    """
    if "uptime.checks_list" in __salt__:
        return True
    return (False, "uptime module could not be loaded")


def monitored(name, **params):
    """
    Makes sure an URL is monitored by uptime. Checks if URL is already
    monitored, and if not, adds it.

    """

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}
    if __salt__["uptime.check_exists"](name=name):
        ret["result"] = True
        ret["comment"] = "URL {} is already monitored".format(name)
        ret["changes"] = {}
        return ret
    if not __opts__["test"]:
        url_monitored = __salt__["uptime.create"](name, **params)
        if url_monitored:
            ret["result"] = True
            msg = "Successfully added the URL {0} to uptime"
            ret["comment"] = msg.format(name)
            ret["changes"] = {"url_monitored": url_monitored}
        else:
            ret["result"] = False
            ret["comment"] = "Failed to add {} to uptime".format(name)
            ret["changes"] = {}
    else:
        msg = "URL {0} is going to be added to uptime"
        ret.update(result=None, comment=msg.format(name))
    return ret



"""
Beacon to monitor certificate expiration dates from files on the filesystem.

.. versionadded:: 3000

:maintainer: <devops@eitr.tech>
:maturity: new
:depends: OpenSSL
"""
import logging
from datetime import datetime

import salt.utils.beacons
import salt.utils.files

try:
    from OpenSSL import crypto

    HAS_OPENSSL = True
except ImportError:
    HAS_OPENSSL = False

log = logging.getLogger(__name__)

DEFAULT_NOTIFY_DAYS = 45

__virtualname__ = "cert_info"


def __virtual__():
    if HAS_OPENSSL is False:
        err_msg = "OpenSSL library is missing."
        log.error("Unable to load %s beacon: %s", __virtualname__, err_msg)
        return False, err_msg

    return __virtualname__


def validate(config):
    """
    Validate the beacon configuration
    """
    # Configuration for cert_info beacon should be a list of dicts
    if not isinstance(config, list):
        return False, "Configuration for cert_info beacon must be a list."

    config = salt.utils.beacons.list_to_dict(config)

    if "files" not in config:
        return (
            False,
            "Configuration for cert_info beacon must contain files option.",
        )
    return True, "Valid beacon configuration"


def beacon(config):
    """
    Monitor the certificate files on the minion.

    Specify a notification threshold in days and only emit a beacon if any certificates are
    expiring within that timeframe or if `notify_days` equals `-1` (always report information).
    The default notification threshold is 45 days and can be overridden at the beacon level and
    at an individual certificate level.

    .. code-block:: yaml

        beacons:
          cert_info:
            - files:
                - /etc/pki/tls/certs/mycert.pem
                - /etc/pki/tls/certs/yourcert.pem:
                    notify_days: 15
                - /etc/pki/tls/certs/ourcert.pem
            - notify_days: 45
            - interval: 86400

    """
    ret = []
    certificates = []
    CryptoError = crypto.Error  # pylint: disable=invalid-name

    config = salt.utils.beacons.list_to_dict(config)

    global_notify_days = config.get("notify_days", DEFAULT_NOTIFY_DAYS)

    for cert_path in config.get("files", []):
        notify_days = global_notify_days

        if isinstance(cert_path, dict):
            try:
                next_cert_path = next(iter(cert_path))
                notify_days = cert_path[next_cert_path].get(
                    "notify_days", global_notify_days
                )
            except StopIteration as exc:
                log.error("Unable to load certificate %s (%s)", cert_path, exc)
                continue
            else:
                cert_path = next_cert_path

        try:
            with salt.utils.files.fopen(cert_path) as fp_:
                cert = crypto.load_certificate(crypto.FILETYPE_PEM, fp_.read())
        except (OSError, CryptoError) as exc:
            log.error("Unable to load certificate %s (%s)", cert_path, exc)
            continue

        cert_date = datetime.strptime(
            cert.get_notAfter().decode(encoding="UTF-8"), "%Y%m%d%H%M%SZ"
        )
        date_diff = (cert_date - datetime.today()).days
        log.debug("Certificate %s expires in %s days.", cert_path, date_diff)

        if notify_days < 0 or date_diff <= notify_days:
            log.debug(
                "Certificate %s triggered beacon due to %s day notification threshold.",
                cert_path,
                notify_days,
            )
            extensions = []
            for ext in range(0, cert.get_extension_count()):
                extensions.append(
                    {
                        "ext_name": cert.get_extension(ext)
                        .get_short_name()
                        .decode(encoding="UTF-8"),
                        "ext_data": str(cert.get_extension(ext)),
                    }
                )

            certificates.append(
                {
                    "cert_path": cert_path,
                    "issuer": ",".join(
                        [
                            '{}="{}"'.format(
                                t[0].decode(encoding="UTF-8"),
                                t[1].decode(encoding="UTF-8"),
                            )
                            for t in cert.get_issuer().get_components()
                        ]
                    ),
                    "issuer_dict": {
                        k.decode("UTF-8"): v.decode("UTF-8")
                        for k, v in cert.get_issuer().get_components()
                    },
                    "notAfter_raw": cert.get_notAfter().decode(encoding="UTF-8"),
                    "notAfter": cert_date.strftime("%Y-%m-%d %H:%M:%SZ"),
                    "notBefore_raw": cert.get_notBefore().decode(encoding="UTF-8"),
                    "notBefore": datetime.strptime(
                        cert.get_notBefore().decode(encoding="UTF-8"), "%Y%m%d%H%M%SZ"
                    ).strftime("%Y-%m-%d %H:%M:%SZ"),
                    "serial_number": cert.get_serial_number(),
                    "signature_algorithm": cert.get_signature_algorithm().decode(
                        encoding="UTF-8"
                    ),
                    "subject": ",".join(
                        [
                            '{}="{}"'.format(
                                t[0].decode(encoding="UTF-8"),
                                t[1].decode(encoding="UTF-8"),
                            )
                            for t in cert.get_subject().get_components()
                        ]
                    ),
                    "subject_dict": {
                        k.decode("UTF-8"): v.decode("UTF-8")
                        for k, v in cert.get_subject().get_components()
                    },
                    "version": cert.get_version(),
                    "extensions": extensions,
                    "has_expired": cert.has_expired(),
                }
            )

    if certificates:
        ret.append({"certificates": certificates})

    return ret



import pytest

import salt.utils.json
from salt.netapi.rest_tornado import saltnado


@pytest.fixture
def app_urls():
    return [
        ("/run", saltnado.RunSaltAPIHandler),
    ]


@pytest.mark.parametrize("client", ["local", "local_async", "runner", "runner_async"])
async def test_authentication_exception_consistency(
    http_client, client, content_type_map
):
    """
    Test consistency of authentication exception of each clients.
    """
    valid_response = {"return": ["Failed to authenticate"]}

    request_lowstate = {
        "client": client,
        "tgt": "*",
        "fun": "test.fib",
        "arg": ["10"],
    }

    response = await http_client.fetch(
        "/run",
        method="POST",
        body=salt.utils.json.dumps(request_lowstate),
        headers={"Content-Type": content_type_map["json"]},
    )

    assert salt.utils.json.loads(response.body) == valid_response



import os

import salt.modules.dpkg_lowpkg as dpkg
from tests.support.mock import MagicMock, mock_open, patch


def test_get_pkg_license():
    """
    Test _get_pkg_license for ignore errors on reading license from copyright files
    """
    license_read_mock = mock_open(read_data="")
    with patch.object(os.path, "exists", MagicMock(return_value=True)), patch(
        "salt.utils.files.fopen", license_read_mock
    ):
        dpkg._get_pkg_license("bash")

        assert license_read_mock.calls[0].args[0] == "/usr/share/doc/bash/copyright"
        assert license_read_mock.calls[0].kwargs["errors"] == "ignore"



"""
Tests for the rabbitmq state
"""

import pytest

from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin


@pytest.mark.skip_if_not_root
class RabbitVHostTestCase(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the rabbitmq virtual host states.
    """

    def setUp(self):
        super().setUp()
        rabbit_installed = self.run_function("cmd.has_exec", ["rabbitmqctl"])

        if not rabbit_installed:
            self.skipTest("rabbitmq-server not installed")

    def test_present(self):
        """
        rabbitmq_vhost.present null_host
        """
        ret = self.run_state("rabbitmq_vhost.present", name="null_host", test=True)
        self.assertSaltFalseReturn(ret)

    def absent(self):
        """
        rabbitmq_vhost.absent null_host
        """
        ret = self.run_state("rabbitmq_vhost.absent", name="null_host", test=True)
        self.assertSaltFalseReturn(ret)



import os

import pytest

import salt.utils.verify
from tests.support.mock import patch


@pytest.mark.skip_on_windows(reason="Not applicable for Windows.")
def test_verify_env_race_condition():
    def _stat(path):
        """
        Helper function for mock_stat, we want to raise errors for specific paths, but not until we get into the proper path.
        Until then, just return plain os.stat_result
        """
        if path in ("/tmp/salt-dir/.file3", "/tmp/salt-dir/.dir3"):
            raise AssertionError("The .file3 and .dir3 paths should never be called!")

        if path in ("/tmp/salt-dir/file1", "/tmp/salt-dir/dir1"):
            raise FileNotFoundError(
                "[Errno 2] No such file or directory: this exception should not be visible"
            )

        # we need to return at least different st_uid in order to trigger chown for these paths
        if path in ("/tmp/salt-dir/file4", "/tmp/salt-dir/dir4"):
            return os.stat_result([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])

        return os.stat_result([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

    def _chown(path, uid, gid):
        if path in ("/tmp/salt-dir/file4", "/tmp/salt-dir/dir4"):
            raise FileNotFoundError(
                "[Errno 2] No such file or directory: this exception should not be visible"
            )

        return

    with patch("os.chown", side_effect=_chown) as mock_chown, patch(
        "os.stat", side_effect=_stat
    ) as mock_stat, patch(
        "salt.utils.verify._get_pwnam", return_value=(None, None, 0, 0)
    ), patch(
        "os.getuid", return_value=0
    ), patch(
        "os.listdir", return_value=["subdir"]
    ), patch(
        "os.path.isdir", return_value=True
    ), patch(
        "salt.utils.path.os_walk",
        return_value=[
            (
                "/tmp/salt-dir",
                ["dir1", "dir2", ".dir3", "dir4"],
                ["file1", "file2", ".file3", "file4"],
            )
        ],
    ):

        # verify this runs without issues, even though FNFE is raised
        salt.utils.verify.verify_env(["/tmp/salt-dir"], "root", skip_extra=True)

        # and verify it got actually called with the valid paths
        mock_stat.assert_any_call("/tmp/salt-dir/file1")
        mock_stat.assert_any_call("/tmp/salt-dir/dir1")

        mock_stat.assert_any_call("/tmp/salt-dir/file4")
        mock_stat.assert_any_call("/tmp/salt-dir/dir4")

        mock_chown.assert_any_call("/tmp/salt-dir/file4", 0, 0)
        mock_chown.assert_any_call("/tmp/salt-dir/dir4", 0, 0)



import pytest

import salt.modules.saltutil as saltutil
import salt.states.saltmod as saltmod
from tests.support.mock import create_autospec, patch


@pytest.fixture
def configure_loader_modules():
    return {saltmod: {"__opts__": {"__role": "testsuite"}}}


@pytest.fixture
def fake_cmd():
    fake_cmd = create_autospec(saltutil.cmd)
    with patch.dict(saltmod.__salt__, {"saltutil.cmd": fake_cmd}):
        yield fake_cmd


@pytest.mark.parametrize(
    "exclude",
    [True, False],
)
def test_exclude_parameter_gets_passed(exclude, fake_cmd):
    """
    Smoke test for for salt.states.statemod.state().  Ensures that we
    don't take an exception if optional parameters are not specified in
    __opts__ or __env__.
    """
    args = ("webserver_setup", "webserver2")
    expected_exclude = exclude
    kwargs = {
        "tgt_type": "glob",
        "exclude": expected_exclude,
        "highstate": True,
    }

    saltmod.state(*args, **kwargs)

    call = fake_cmd.call_args[1]
    assert call["kwarg"]["exclude"] == expected_exclude


def test_exclude_parameter_is_not_passed_if_not_provided(fake_cmd):
    # Make sure we don't barf on existing behavior
    args = ("webserver_setup", "webserver2")
    kwargs_without_exclude = {
        "tgt_type": "glob",
        "highstate": True,
    }

    saltmod.state(*args, **kwargs_without_exclude)

    call = fake_cmd.call_args[1]
    assert "exclude" not in call["kwarg"]



"""
Test the win_wua execution module
"""
import pytest

import salt.modules.win_wua as win_wua
import salt.utils.platform
import salt.utils.win_update
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def updates_list():
    return {
        "ca3bb521-a8ea-4e26-a563-2ad6e3108b9a": {"KBs": ["KB4481252"]},
        "07609d43-d518-4e77-856e-d1b316d1b8a8": {"KBs": ["KB925673"]},
        "fbaa5360-a440-49d8-a3b6-0c4fc7ecaa19": {"KBs": ["KB4481252"]},
        "a873372b-7a5c-443c-8022-cd59a550bef4": {"KBs": ["KB3193497"]},
        "14075cbe-822e-4004-963b-f50e08d45563": {"KBs": ["KB4540723"]},
        "d931e99c-4dda-4d39-9905-0f6a73f7195f": {"KBs": ["KB3193497"]},
        "afda9e11-44a0-4602-9e9b-423af11ecaed": {"KBs": ["KB4541329"]},
        "a0f997b1-1abe-4a46-941f-b37f732f9fbd": {"KBs": ["KB3193497"]},
        "eac02b09-d745-4891-b80f-400e0e5e4b6d": {"KBs": ["KB4052623"]},
        "0689e74b-54d1-4f55-a916-96e3c737db90": {"KBs": ["KB890830"]},
    }


@pytest.fixture
def updates_summary():
    return {"Installed": 10}


class Updates:
    @staticmethod
    def list():
        return {
            "ca3bb521-a8ea-4e26-a563-2ad6e3108b9a": {"KBs": ["KB4481252"]},
            "07609d43-d518-4e77-856e-d1b316d1b8a8": {"KBs": ["KB925673"]},
            "fbaa5360-a440-49d8-a3b6-0c4fc7ecaa19": {"KBs": ["KB4481252"]},
            "a873372b-7a5c-443c-8022-cd59a550bef4": {"KBs": ["KB3193497"]},
            "14075cbe-822e-4004-963b-f50e08d45563": {"KBs": ["KB4540723"]},
            "d931e99c-4dda-4d39-9905-0f6a73f7195f": {"KBs": ["KB3193497"]},
            "afda9e11-44a0-4602-9e9b-423af11ecaed": {"KBs": ["KB4541329"]},
            "a0f997b1-1abe-4a46-941f-b37f732f9fbd": {"KBs": ["KB3193497"]},
            "eac02b09-d745-4891-b80f-400e0e5e4b6d": {"KBs": ["KB4052623"]},
            "0689e74b-54d1-4f55-a916-96e3c737db90": {"KBs": ["KB890830"]},
        }

    @staticmethod
    def summary():
        return {"Installed": 10}


def test__virtual__not_windows():
    """
    Test __virtual__ function on Non-Windows
    """
    with patch("salt.utils.platform.is_windows", autospec=True, return_value=False):
        expected = (False, "WUA: Only available on Windows systems")
        result = win_wua.__virtual__()
        assert result == expected


def test__virtual__missing_pywin32():
    """
    Test __virtual__ function when pywin32 is not installed
    """
    with patch("salt.modules.win_wua.HAS_PYWIN32", False):
        expected = (False, "WUA: Requires PyWin32 libraries")
        result = win_wua.__virtual__()
        assert result == expected


def test__virtual__wuauserv_disabled():
    """
    Test __virtual__ function when the wuauserv service is disabled
    """
    mock_service_info = MagicMock(
        side_effect=[
            {"StartType": "Disabled"},  # wuauserv
            {"StartType": "Auto"},  # msiserver
            {"StartType": "Auto"},  # BITS
            {"StartType": "Auto"},  # CryptSvc
            {"StartType": "Auto"},  # TrustedInstaller
        ]
    )
    with patch("salt.utils.win_service.info", mock_service_info):
        expected = (
            False,
            "WUA: The Windows Update service (wuauserv) must not be disabled",
        )
        result = win_wua.__virtual__()
        assert result == expected


def test__virtual__msiserver_disabled():
    """
    Test __virtual__ function when the msiserver service is disabled
    """
    mock_service_info = MagicMock(
        side_effect=[
            {"StartType": "Auto"},  # wuauserv
            {"StartType": "Disabled"},  # msiserver
            {"StartType": "Auto"},  # BITS
            {"StartType": "Auto"},  # CryptSvc
            {"StartType": "Auto"},  # TrustedInstaller
        ]
    )
    with patch("salt.utils.win_service.info", mock_service_info):
        expected = (
            False,
            "WUA: The Windows Installer service (msiserver) must not be disabled",
        )
        result = win_wua.__virtual__()
        assert result == expected


def test__virtual__BITS_disabled():
    """
    Test __virtual__ function when the BITS service is disabled
    """
    mock_service_info = MagicMock(
        side_effect=[
            {"StartType": "Auto"},  # wuauserv
            {"StartType": "Auto"},  # msiserver
            {"StartType": "Disabled"},  # BITS
            {"StartType": "Auto"},  # CryptSvc
            {"StartType": "Auto"},  # TrustedInstaller
        ]
    )
    with patch("salt.utils.win_service.info", mock_service_info):
        expected = (
            False,
            "WUA: The Background Intelligent Transfer service (bits) must not be"
            " disabled",
        )
        result = win_wua.__virtual__()
        assert result == expected


def test__virtual__BITS_manual():
    """
    Test __virtual__ function when the BITS service is set to manual
    Should not disable the module (__virtual__ should return True)
    """
    mock_service_info = MagicMock(
        side_effect=[
            {"StartType": "Auto"},  # wuauserv
            {"StartType": "Auto"},  # msiserver
            {"StartType": "Manual"},  # BITS
            {"StartType": "Auto"},  # CryptSvc
            {"StartType": "Auto"},  # TrustedInstaller
        ]
    )
    with patch("salt.utils.win_service.info", mock_service_info):
        expected = True
        result = win_wua.__virtual__()
        assert result == expected


def test__virtual__CryptSvc_disabled():
    """
    Test __virtual__ function when the CryptSvc service is disabled
    """
    mock_service_info = MagicMock(
        side_effect=[
            {"StartType": "Auto"},  # wuauserv
            {"StartType": "Auto"},  # msiserver
            {"StartType": "Auto"},  # BITS
            {"StartType": "Disabled"},  # CryptSvc
            {"StartType": "Auto"},  # TrustedInstaller
        ]
    )
    with patch("salt.utils.win_service.info", mock_service_info):
        expected = (
            False,
            "WUA: The Cryptographic Services service (CryptSvc) must not be"
            " disabled",
        )
        result = win_wua.__virtual__()
        assert result == expected


def test__virtual__CryptSvc_manual():
    """
    Test __virtual__ function when the CryptSvc service is disabled
    """
    mock_service_info = MagicMock(
        side_effect=[
            {"StartType": "Auto"},  # wuauserv
            {"StartType": "Auto"},  # msiserver
            {"StartType": "Auto"},  # BITS
            {"StartType": "Manual"},  # CryptSvc
            {"StartType": "Auto"},  # TrustedInstaller
        ]
    )
    with patch("salt.utils.win_service.info", mock_service_info):
        result = win_wua.__virtual__()
        assert result is True


def test__virtual__TrustedInstaller_disabled():
    """
    Test __virtual__ function when the TrustedInstaller service is disabled
    """
    mock_service_info = MagicMock(
        side_effect=[
            {"StartType": "Auto"},  # wuauserv
            {"StartType": "Auto"},  # msiserver
            {"StartType": "Auto"},  # BITS
            {"StartType": "Auto"},  # CryptSvc
            {"StartType": "Disabled"},  # TrustedInstaller
        ]
    )
    with patch("salt.utils.win_service.info", mock_service_info):
        expected = (
            False,
            "WUA: The Windows Module Installer service (TrustedInstaller) must not"
            " be disabled",
        )
        result = win_wua.__virtual__()
        assert result == expected


def test_installed(updates_list):
    """
    Test installed function default
    """
    expected = updates_list
    with patch("salt.utils.winapi.Com", autospec=True), patch(
        "win32com.client.Dispatch", autospec=True
    ), patch.object(
        salt.utils.win_update.WindowsUpdateAgent, "refresh", autospec=True
    ), patch.object(
        salt.utils.win_update, "Updates", autospec=True, return_value=Updates()
    ):
        result = win_wua.installed()
        assert result == expected


def test_installed_summary(updates_summary):
    """
    Test installed function with summary=True
    """
    expected = updates_summary
    # Remove all updates that are not installed
    with patch("salt.utils.winapi.Com", autospec=True), patch(
        "win32com.client.Dispatch", autospec=True
    ), patch.object(
        salt.utils.win_update.WindowsUpdateAgent, "refresh", autospec=True
    ), patch.object(
        salt.utils.win_update, "Updates", autospec=True, return_value=Updates()
    ):
        result = win_wua.installed(summary=True)
        assert result == expected


def test_installed_kbs_only(updates_list):
    """
    Test installed function with kbs_only=True
    """
    expected = set()
    for update in updates_list:
        expected.update(updates_list[update]["KBs"])
    expected = sorted(expected)
    # Remove all updates that are not installed
    with patch("salt.utils.winapi.Com", autospec=True), patch(
        "win32com.client.Dispatch", autospec=True
    ), patch.object(
        salt.utils.win_update.WindowsUpdateAgent, "refresh", autospec=True
    ), patch.object(
        salt.utils.win_update, "Updates", autospec=True, return_value=Updates()
    ):
        result = win_wua.installed(kbs_only=True)
        assert result == expected



"""
BGP Finder
==========

.. versionadded:: 2017.7.0

Runner to search BGP neighbors details.

Configuration
-------------

- Minion (proxy) config

    The ``bgp.neighbors`` function must be appened in the list of ``mine_functions``:

    .. code-block:: yaml

        mine_functions:
          bgp.neighbors: []

    Which instructs Salt to cache the data returned by the ``neighbors`` function
    from the :mod:`NAPALM BGP module <salt.modules.napalm_bgp.neighbors>`.

    How often the mines are refreshed, can be specified using:

    .. code-block:: yaml

        mine_interval: <X minutes>

- Master config

    By default the following options can be configured on the master.
    They are not mandatory, but available in case the user has different requirements.

    tgt: ``*``
        From what minions will collect the mine data.
        Default: ``*`` (collect mine data from all minions)

    tgt_type: ``glob``
        Minion matching expression form. Default: ``glob``.

    return_fields
        What fields to return in the output.
        It can display all the fields from the ``neighbors`` function
        from the :mod:`NAPALM BGP module <salt.modules.napalm_bgp.neighbors>`.

        Some fields cannot be removed:

        - ``as_number``: the AS number of the neighbor
        - ``device``: the minion ID
        - ``neighbor_address``: the neighbor remote IP address

        By default, the following extra fields are returned (displayed):

        - ``connection_stats``: connection stats, as described below
        - ``import_policy``: the name of the import policy
        - ``export_policy``: the name of the export policy

        Special fields:

        - ``vrf``: return the name of the VRF.
        - ``connection_stats``: returning an output of the form ``<State>
          <Active>/<Received>/<Accepted>/<Damped>``, e.g.  ``Established
          398/399/399/0`` similar to the usual output from network devices.
        - ``interface_description``: matches the neighbor details with the
          corresponding interface and returns its description. This will reuse
          functionality from the :mod:`net runner
          <salt.runners.net.interfaces>`, so the user needs to enable the mines
          as specified in the documentation.
        - ``interface_name``: matches the neighbor details with the
          corresponding interface and returns the name.  Similar to
          ``interface_description``, this will reuse functionality from the
          :mod:`net runner <salt.runners.net.interfaces>`, so the user needs to
          enable the mines as specified in the documentation.

    display: ``True``
        Display on the screen or return structured object? Default: ``True`` (return on the CLI).

    outputter: ``table``
        Specify the outputter name when displaying on the CLI. Default: :mod:`table <salt.output.table_out>`.

    Configuration example:

    .. code-block:: yaml

        runners:
          bgp:
            tgt: 'edge*'
            tgt_type: 'glob'
            return_fields:
              - up
              - connection_state
              - previous_connection_state
              - suppress_4byte_as
              - holdtime
              - flap_count
            outputter: yaml
"""

import salt.output

try:
    # pylint: disable=unused-import,no-name-in-module
    from napalm.base import helpers as napalm_helpers
    from netaddr import IPAddress, IPNetwork

    # pylint: enable=unused-import,no-name-in-module

    HAS_NAPALM = True
except ImportError:
    HAS_NAPALM = False


# -----------------------------------------------------------------------------
# module properties
# -----------------------------------------------------------------------------

__virtualname__ = "bgp"

# -----------------------------------------------------------------------------
# global variables
# -----------------------------------------------------------------------------

_DEFAULT_TARGET = "*"
_DEFAULT_EXPR_FORM = "glob"
_DEFAULT_DISPLAY = True
_DEFAULT_OUTPUTTER = "table"
_DEFAULT_INCLUDED_FIELDS = ["device", "as_number", "neighbor_address"]
_DEFAULT_RETURN_FIELDS = ["connection_stats", "import_policy", "export_policy"]
_DEFAULT_LABELS_MAPPING = {
    "device": "Device",
    "as_number": "AS Number",
    "neighbor_address": "Neighbor IP",
    "connection_stats": "State|#Active/Received/Accepted/Damped",
    "import_policy": "Policy IN",
    "export_policy": "Policy OUT",
    "vrf": "VRF",
}

# -----------------------------------------------------------------------------
# property functions
# -----------------------------------------------------------------------------


def __virtual__():
    if HAS_NAPALM:
        return __virtualname__
    return (False, "The napalm module could not be imported")


# -----------------------------------------------------------------------------
# helper functions -- will not be exported
# -----------------------------------------------------------------------------


def _get_bgp_runner_opts():
    """
    Return the bgp runner options.
    """
    runner_opts = __opts__.get("runners", {}).get("bgp", {})
    return {
        "tgt": runner_opts.get("tgt", _DEFAULT_TARGET),
        "tgt_type": runner_opts.get("tgt_type", _DEFAULT_EXPR_FORM),
        "display": runner_opts.get("display", _DEFAULT_DISPLAY),
        "return_fields": _DEFAULT_INCLUDED_FIELDS
        + runner_opts.get("return_fields", _DEFAULT_RETURN_FIELDS),
        "outputter": runner_opts.get("outputter", _DEFAULT_OUTPUTTER),
    }


def _get_mine(opts=None):
    """
    Helper to return the mine data from the minions, as configured on the runner opts.
    """
    if not opts:
        # not a massive improvement, but better than recomputing the runner opts dict
        opts = _get_bgp_runner_opts()
    return __salt__["mine.get"](opts["tgt"], "bgp.neighbors", tgt_type=opts["tgt_type"])


def _compare_match(dict1, dict2):
    """
    Compare two dictionaries and return a boolean value if their values match.
    """
    for karg, warg in dict1.items():
        if karg in dict2 and dict2[karg] != warg:
            return False
    return True


def _display_runner(
    rows, labels, title, display=_DEFAULT_DISPLAY, outputter=_DEFAULT_OUTPUTTER
):
    """
    Display or return the rows.
    """
    if display:
        if outputter == "table":
            ret = salt.output.out_format(
                {"rows": rows, "labels": labels},
                "table",
                __opts__,
                title=title,
                rows_key="rows",
                labels_key="labels",
            )
        else:
            ret = salt.output.out_format(rows, outputter, __opts__)
        print(ret)
    else:
        return rows


# -----------------------------------------------------------------------------
# callable functions
# -----------------------------------------------------------------------------


def neighbors(*asns, **kwargs):
    """
    Search for BGP neighbors details in the mines of the ``bgp.neighbors`` function.

    Arguments:

    asns
        A list of AS numbers to search for.
        The runner will return only the neighbors of these AS numbers.

    device
        Filter by device name (minion ID).

    ip
        Search BGP neighbor using the IP address.
        In multi-VRF environments, the same IP address could be used by
        more than one neighbors, in different routing tables.

    network
        Search neighbors within a certain IP network.

    title
        Custom title.

    display: ``True``
        Display on the screen or return structured object? Default: ``True`` (return on the CLI).

    outputter: ``table``
        Specify the outputter name when displaying on the CLI. Default: :mod:`table <salt.output.table_out>`.

    In addition, any field from the output of the ``neighbors`` function
    from the :mod:`NAPALM BGP module <salt.modules.napalm_bgp.neighbors>` can be used as a filter.

    CLI Example:

    .. code-block:: bash

        salt-run bgp.neighbors 13335 15169
        salt-run bgp.neighbors 13335 ip=172.17.19.1
        salt-run bgp.neighbors multipath=True
        salt-run bgp.neighbors up=False export_policy=my-export-policy multihop=False
        salt-run bgp.neighbors network=192.168.0.0/16

    Output example:

    .. code-block:: text

        BGP Neighbors for 13335, 15169
        ________________________________________________________________________________________________________________________________________________________________
        |    Device    | AS Number |         Neighbor Address        | State|#Active/Received/Accepted/Damped |         Policy IN         |         Policy OUT         |
        ________________________________________________________________________________________________________________________________________________________________
        | edge01.bjm01 |   13335   |          172.17.109.11          |        Established 0/398/398/0         |       import-policy       |        export-policy       |
        ________________________________________________________________________________________________________________________________________________________________
        | edge01.bjm01 |   13335   |          172.17.109.12          |       Established 397/398/398/0        |       import-policy       |        export-policy       |
        ________________________________________________________________________________________________________________________________________________________________
        | edge01.flw01 |   13335   |          192.168.172.11         |        Established 1/398/398/0         |       import-policy       |        export-policy       |
        ________________________________________________________________________________________________________________________________________________________________
        | edge01.oua01 |   13335   |          172.17.109.17          |          Established 0/0/0/0           |       import-policy       |        export-policy       |
        ________________________________________________________________________________________________________________________________________________________________
        | edge01.bjm01 |   15169   |             2001::1             |       Established 102/102/102/0        |       import-policy       |        export-policy       |
        ________________________________________________________________________________________________________________________________________________________________
        | edge01.bjm01 |   15169   |             2001::2             |       Established 102/102/102/0        |       import-policy       |        export-policy       |
        ________________________________________________________________________________________________________________________________________________________________
        | edge01.tbg01 |   13335   |          192.168.172.17         |          Established 0/1/1/0           |       import-policy       |        export-policy       |
        ________________________________________________________________________________________________________________________________________________________________
    """
    opts = _get_bgp_runner_opts()
    title = kwargs.pop("title", None)
    display = kwargs.pop("display", opts["display"])
    outputter = kwargs.pop("outputter", opts["outputter"])

    # cleaning up the kwargs
    # __pub args not used in this runner (yet)
    kwargs_copy = {}
    kwargs_copy.update(kwargs)
    for karg, _ in kwargs_copy.items():
        if karg.startswith("__pub"):
            kwargs.pop(karg)
    if not asns and not kwargs:
        if display:
            print("Please specify at least an AS Number or an output filter")
        return []
    device = kwargs.pop("device", None)
    neighbor_ip = kwargs.pop("ip", None)
    ipnet = kwargs.pop("network", None)
    ipnet_obj = IPNetwork(ipnet) if ipnet else None
    # any other key passed on the CLI can be used as a filter

    rows = []
    # building the labels
    labels = {}
    for field in opts["return_fields"]:
        if field in _DEFAULT_LABELS_MAPPING:
            labels[field] = _DEFAULT_LABELS_MAPPING[field]
        else:
            # transform from 'previous_connection_state' to 'Previous Connection State'
            labels[field] = " ".join(map(lambda word: word.title(), field.split("_")))
    display_fields = list(set(opts["return_fields"]) - set(_DEFAULT_INCLUDED_FIELDS))
    get_bgp_neighbors_all = _get_mine(opts=opts)

    if not title:
        title_parts = []
        if asns:
            title_parts.append(
                "BGP Neighbors for {asns}".format(
                    asns=", ".join([str(asn) for asn in asns])
                )
            )
        if neighbor_ip:
            title_parts.append(
                "Selecting neighbors having the remote IP address: {ipaddr}".format(
                    ipaddr=neighbor_ip
                )
            )
        if ipnet:
            title_parts.append(
                "Selecting neighbors within the IP network: {ipnet}".format(ipnet=ipnet)
            )
        if kwargs:
            title_parts.append(
                "Searching for BGP neighbors having the attributes: {attrmap}".format(
                    attrmap=", ".join(
                        map(
                            lambda key: "{key}={value}".format(
                                key=key, value=kwargs[key]
                            ),
                            kwargs,
                        )
                    )
                )
            )
        title = "\n".join(title_parts)
    for (
        minion,
        get_bgp_neighbors_minion,
    ) in get_bgp_neighbors_all.items():  # pylint: disable=too-many-nested-blocks
        if not get_bgp_neighbors_minion.get("result"):
            continue  # ignore empty or failed mines
        if device and minion != device:
            # when requested to display only the neighbors on a certain device
            continue
        get_bgp_neighbors_minion_out = get_bgp_neighbors_minion.get("out", {})
        for (
            vrf,
            vrf_bgp_neighbors,
        ) in get_bgp_neighbors_minion_out.items():  # pylint: disable=unused-variable
            for asn, get_bgp_neighbors_minion_asn in vrf_bgp_neighbors.items():
                if asns and asn not in asns:
                    # if filtering by AS number(s),
                    # will ignore if this AS number key not in that list
                    # and continue the search
                    continue
                for neighbor in get_bgp_neighbors_minion_asn:
                    if kwargs and not _compare_match(kwargs, neighbor):
                        # requested filtering by neighbors stats
                        # but this one does not correspond
                        continue
                    if neighbor_ip and neighbor_ip != neighbor.get("remote_address"):
                        # requested filtering by neighbors IP addr
                        continue
                    if ipnet_obj and neighbor.get("remote_address"):
                        neighbor_ip_obj = IPAddress(neighbor.get("remote_address"))
                        if neighbor_ip_obj not in ipnet_obj:
                            # Neighbor not in this network
                            continue
                    row = {
                        "device": minion,
                        "neighbor_address": neighbor.get("remote_address"),
                        "as_number": asn,
                    }
                    if "vrf" in display_fields:
                        row["vrf"] = vrf
                    if "connection_stats" in display_fields:
                        connection_stats = (
                            "{state} {active}/{received}/{accepted}/{damped}".format(
                                state=neighbor.get("connection_state", -1),
                                active=neighbor.get("active_prefix_count", -1),
                                received=neighbor.get("received_prefix_count", -1),
                                accepted=neighbor.get("accepted_prefix_count", -1),
                                damped=neighbor.get("suppressed_prefix_count", -1),
                            )
                        )
                        row["connection_stats"] = connection_stats
                    if (
                        "interface_description" in display_fields
                        or "interface_name" in display_fields
                    ):
                        net_find = __salt__["net.interfaces"](
                            device=minion,
                            ipnet=neighbor.get("remote_address"),
                            display=False,
                        )
                        if net_find:
                            if "interface_description" in display_fields:
                                row["interface_description"] = net_find[0][
                                    "interface_description"
                                ]
                            if "interface_name" in display_fields:
                                row["interface_name"] = net_find[0]["interface"]
                        else:
                            # if unable to find anything, leave blank
                            if "interface_description" in display_fields:
                                row["interface_description"] = ""
                            if "interface_name" in display_fields:
                                row["interface_name"] = ""
                    for field in display_fields:
                        if field in neighbor:
                            row[field] = neighbor[field]
                    rows.append(row)
    return _display_runner(rows, labels, title, display=display, outputter=outputter)



import importlib

import pytest

import salt.exceptions
from tests.support.mock import MagicMock, patch

# "pass" is a reserved keyword, we need to import it differently
pass_ = importlib.import_module("salt.renderers.pass")


@pytest.fixture
def configure_loader_modules(master_opts):
    return {
        pass_: {
            "__opts__": master_opts,
            "_get_pass_exec": MagicMock(return_value="/usr/bin/pass"),
        }
    }


# The default behavior is that if fetching a secret from pass fails,
# the value is passed through. Even the trailing newlines are preserved.
def test_passthrough():
    pass_path = "secret\n"
    expected = pass_path
    result = pass_.render(pass_path)

    assert result == expected


# Fetch a secret in the strict mode.
def test_strict_fetch():
    config = {
        "pass_variable_prefix": "pass:",
        "pass_strict_fetch": True,
    }

    popen_mock = MagicMock(spec=pass_.Popen)
    popen_mock.return_value.communicate.return_value = ("password123456\n", "")
    popen_mock.return_value.returncode = 0

    mocks = {
        "Popen": popen_mock,
    }

    pass_path = "pass:secret"
    expected = "password123456"
    with patch.dict(pass_.__opts__, config), patch.dict(pass_.__dict__, mocks):
        result = pass_.render(pass_path)

    assert result == expected


# Fail to fetch a secret in the strict mode.
def test_strict_fetch_fail():
    config = {
        "pass_variable_prefix": "pass:",
        "pass_strict_fetch": True,
    }

    popen_mock = MagicMock(spec=pass_.Popen)
    popen_mock.return_value.communicate.return_value = ("", "Secret not found")
    popen_mock.return_value.returncode = 1

    mocks = {
        "Popen": popen_mock,
    }

    pass_path = "pass:secret"
    with patch.dict(pass_.__opts__, config), patch.dict(pass_.__dict__, mocks):
        with pytest.raises(salt.exceptions.SaltRenderError):
            pass_.render(pass_path)


# Passthrough a value that doesn't have a pass prefix.
def test_strict_fetch_passthrough():
    config = {
        "pass_variable_prefix": "pass:",
        "pass_strict_fetch": True,
    }

    pass_path = "variable-without-pass-prefix\n"
    expected = pass_path
    with patch.dict(pass_.__opts__, config):
        result = pass_.render(pass_path)

    assert result == expected


# Fetch a secret in the strict mode. The pass path contains spaces.
def test_strict_fetch_pass_path_with_spaces():
    config = {
        "pass_variable_prefix": "pass:",
        "pass_strict_fetch": True,
    }

    popen_mock = MagicMock(spec=pass_.Popen)
    popen_mock.return_value.communicate.return_value = ("password123456\n", "")
    popen_mock.return_value.returncode = 0

    mocks = {
        "Popen": popen_mock,
    }

    pass_path = "pass:se cr et"
    with patch.dict(pass_.__opts__, config), patch.dict(pass_.__dict__, mocks):
        pass_.render(pass_path)

    call_args, call_kwargs = popen_mock.call_args_list[0]
    assert call_args[0] == ["/usr/bin/pass", "show", "se cr et"]


# Fetch a secret in the strict mode. The secret contains leading and trailing whitespaces.
def test_strict_fetch_secret_with_whitespaces():
    config = {
        "pass_variable_prefix": "pass:",
        "pass_strict_fetch": True,
    }

    popen_mock = MagicMock(spec=pass_.Popen)
    popen_mock.return_value.communicate.return_value = (" \tpassword123456\t \r\n", "")
    popen_mock.return_value.returncode = 0

    mocks = {
        "Popen": popen_mock,
    }

    pass_path = "pass:secret"
    expected = " \tpassword123456\t "  # only the trailing newlines get striped
    with patch.dict(pass_.__opts__, config), patch.dict(pass_.__dict__, mocks):
        result = pass_.render(pass_path)

    assert result == expected


# Test setting env variables based on config values:
# - pass_gnupghome -> GNUPGHOME
# - pass_dir -> PASSWORD_STORE_DIR
def test_env():
    config = {
        "pass_variable_prefix": "pass:",
        "pass_strict_fetch": True,
        "pass_gnupghome": "/path/to/gnupghome",
        "pass_dir": "/path/to/secretstore",
    }

    popen_mock = MagicMock(spec=pass_.Popen)
    popen_mock.return_value.communicate.return_value = ("password123456\n", "")
    popen_mock.return_value.returncode = 0

    mocks = {
        "Popen": popen_mock,
    }

    pass_path = "pass:secret"
    expected = " \tpassword123456\t "  # only the trailing newlines get striped
    with patch.dict(pass_.__opts__, config), patch.dict(pass_.__dict__, mocks):
        result = pass_.render(pass_path)

    call_args, call_kwargs = popen_mock.call_args_list[0]
    assert call_kwargs["env"]["GNUPGHOME"] == config["pass_gnupghome"]
    assert call_kwargs["env"]["PASSWORD_STORE_DIR"] == config["pass_dir"]



import json
import logging
import threading
from copy import copy

import pytest

import salt.utils.files
import salt.utils.vault as vault
from tests.support.mock import ANY, MagicMock, Mock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def tmp_cache(tmp_path):
    cachedir = tmp_path / "cachedir"
    cachedir.mkdir()
    return cachedir


@pytest.fixture
def configure_loader_modules(tmp_cache):
    return {
        vault: {
            "__opts__": {
                "vault": {
                    "url": "http://127.0.0.1",
                    "auth": {
                        "token": "test",
                        "method": "token",
                        "uses": 15,
                        "ttl": 500,
                    },
                },
                "file_client": "local",
                "cachedir": str(tmp_cache),
            },
            "__grains__": {"id": "test-minion"},
            "__context__": {},
        }
    }


@pytest.fixture
def json_success():
    return {
        "request_id": "35df4df1-c3d8-b270-0682-ddb0160c7450",
        "lease_id": "",
        "renewable": False,
        "lease_duration": 0,
        "data": {
            "data": {"something": "myvalue"},
            "metadata": {
                "created_time": "2020-05-02T07:26:12.180848003Z",
                "deletion_time": "",
                "destroyed": False,
                "version": 1,
            },
        },
        "wrap_info": None,
        "warnings": None,
        "auth": None,
    }


@pytest.fixture
def json_denied():
    return {"errors": ["permission denied"]}


@pytest.fixture
def cache_single():
    return {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": None,
        "uses": 1,
        "lease_duration": 100,
        "issued": 3000,
    }


@pytest.fixture
def cache_single_namespace():
    return {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": "test_namespace",
        "uses": 1,
        "lease_duration": 100,
        "issued": 3000,
    }


@pytest.fixture
def cache_uses():
    return {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": None,
        "uses": 10,
        "lease_duration": 100,
        "issued": 3000,
        "unlimited_use_token": False,
    }


@pytest.fixture
def cache_uses_last():
    return {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": None,
        "uses": 1,
        "lease_duration": 100,
        "issued": 3000,
        "unlimited_use_token": False,
    }


@pytest.fixture
def cache_unlimited():
    return {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": None,
        "uses": 0,
        "lease_duration": 100,
        "issued": 3000,
        "unlimited_use_token": True,
    }


@pytest.fixture
def metadata_v2():
    return {
        "accessor": "kv_f8731f1b",
        "config": {
            "default_lease_ttl": 0,
            "force_no_cache": False,
            "max_lease_ttl": 0,
        },
        "description": "key/value secret storage",
        "external_entropy_access": False,
        "local": False,
        "options": {"version": "2"},
        "path": "secret/",
        "seal_wrap": False,
        "type": "kv",
        "uuid": "1d9431ac-060a-9b63-4572-3ca7ffd78347",
    }


@pytest.fixture
def cache_secret_meta(metadata_v2):
    return {"vault_secret_path_metadata": {"secret/mything": metadata_v2}}


def _mock_json_response(data, status_code=200, reason=""):
    """
    Mock helper for http response
    """
    response = MagicMock()
    response.json = MagicMock(return_value=data)
    response.status_code = status_code
    response.reason = reason
    if status_code == 200:
        response.ok = True
    else:
        response.ok = False
    return Mock(return_value=response)


def test_write_cache_multi_use_token(cache_uses, tmp_cache):
    """
    Test write cache with multi-use token
    """
    expected_write = {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": None,
        "uses": 10,
        "lease_duration": 100,
        "issued": 3000,
        "unlimited_use_token": False,
    }
    function_response = vault.write_cache(cache_uses)
    assert function_response is True
    with salt.utils.files.fopen(str(tmp_cache / "salt_vault_token"), "r") as fp:
        token_data = json.loads(fp.read())
    assert token_data == expected_write


def test_write_cache_unlimited_token(cache_uses, tmp_cache):
    """
    Test write cache with unlimited use token
    """
    write_data = {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": None,
        "uses": 0,
        "lease_duration": 100,
        "issued": 3000,
    }
    expected_write = {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": None,
        "uses": 0,
        "lease_duration": 100,
        "issued": 3000,
        "unlimited_use_token": True,
    }
    function_response = vault.write_cache(write_data)
    with salt.utils.files.fopen(str(tmp_cache / "salt_vault_token"), "r") as fp:
        token_data = json.loads(fp.read())
    assert token_data == expected_write


def test_write_cache_issue_59361(cache_uses, tmp_cache):
    """
    Test race condition fix (Issue 59361)
    """
    evt = threading.Event()

    def target(evt, cache_uses):
        evt.wait()
        function_response = vault.write_cache(cache_uses)

    cached_token = {
        "url": "http://127.0.0.1:8200",
        "token": "testwithmuchmuchlongertoken",
        "verify": None,
        "namespace": None,
        "uses": 10,
        "lease_duration": 100,
        "issued": 3000,
        "unlimited_use_token": False,
    }
    expected_write = {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": None,
        "uses": 10,
        "lease_duration": 100,
        "issued": 3000,
        "unlimited_use_token": False,
    }

    thread1 = threading.Thread(
        target=target,
        args=(
            evt,
            cached_token,
        ),
    )
    thread1.start()
    thread2 = threading.Thread(
        target=target,
        args=(
            evt,
            expected_write,
        ),
    )
    thread2.start()
    evt.set()
    thread1.join()
    thread2.join()

    with salt.utils.files.fopen(str(tmp_cache / "salt_vault_token"), "r") as fp:
        try:
            token_data = json.loads(fp.read())
        except json.decoder.JSONDecodeError:
            assert False, "Cache file data corrupted"


def test_make_request_single_use_token_run_ok(json_success, cache_single):
    """
    Given single use token in __context__, function should run successful secret lookup with no other modifications
    """
    mock = _mock_json_response(json_success)
    supplied_context = {"vault_token": copy(cache_single)}
    expected_headers = {"X-Vault-Token": "test", "Content-Type": "application/json"}
    with patch.dict(vault.__context__, supplied_context):
        with patch("requests.request", mock):
            vault_return = vault.make_request("/secret/my/secret", "key")
            assert vault.__context__ == {}
            mock.assert_called_with(
                "/secret/my/secret",
                "http://127.0.0.1:8200/key",
                headers=expected_headers,
                verify=ANY,
            )
            assert vault_return.json() == json_success


def test_make_request_single_use_token_run_auth_error(json_denied, cache_single):
    """
    Given single use token in __context__ and login error, function should request token and re-run
    """
    # Disable logging because simulated http failures are logged as errors
    logging.disable(logging.CRITICAL)
    mock = _mock_json_response(json_denied, status_code=400)
    supplied_context = {"vault_token": cache_single}
    expected_headers = {"X-Vault-Token": "test", "Content-Type": "application/json"}
    with patch.dict(vault.__context__, supplied_context):
        with patch("requests.request", mock):
            with patch.object(vault, "del_cache") as mock_del_cache:
                vault_return = vault.make_request("/secret/my/secret", "key")
                assert vault.__context__ == {}
                mock.assert_called_with(
                    "/secret/my/secret",
                    "http://127.0.0.1:8200/key",
                    headers=expected_headers,
                    verify=ANY,
                )
                assert vault_return.json() == json_denied
                mock_del_cache.assert_called()
                assert mock.call_count == 2
    logging.disable(logging.NOTSET)


def test_multi_use_token_successful_run(json_success, cache_uses):
    """
    Given multi-use token, function should get secret and decrement token
    """
    expected_cache_write = {
        "url": "http://127.0.0.1:8200",
        "token": "test",
        "verify": None,
        "namespace": None,
        "uses": 9,
        "lease_duration": 100,
        "issued": 3000,
        "unlimited_use_token": False,
    }
    mock = _mock_json_response(json_success)
    expected_headers = {"X-Vault-Token": "test", "Content-Type": "application/json"}
    with patch.object(vault, "get_cache") as mock_get_cache:
        mock_get_cache.return_value = copy(cache_uses)
        with patch("requests.request", mock):
            with patch.object(vault, "del_cache") as mock_del_cache:
                with patch.object(vault, "write_cache") as mock_write_cache:
                    vault_return = vault.make_request("/secret/my/secret", "key")
                    mock.assert_called_with(
                        "/secret/my/secret",
                        "http://127.0.0.1:8200/key",
                        headers=expected_headers,
                        verify=ANY,
                    )
                    mock_write_cache.assert_called_with(expected_cache_write)
                    assert vault_return.json() == json_success
                    assert mock.call_count == 1


def test_multi_use_token_last_use(json_success, cache_uses_last):
    """
    Given last use of multi-use token, function should succeed and flush token cache
    """
    mock = _mock_json_response(json_success)
    expected_headers = {"X-Vault-Token": "test", "Content-Type": "application/json"}
    with patch.object(vault, "get_cache") as mock_get_cache:
        mock_get_cache.return_value = cache_uses_last
        with patch("requests.request", mock):
            with patch.object(vault, "del_cache") as mock_del_cache:
                with patch.object(vault, "write_cache") as mock_write_cache:
                    vault_return = vault.make_request("/secret/my/secret", "key")
                    mock.assert_called_with(
                        "/secret/my/secret",
                        "http://127.0.0.1:8200/key",
                        headers=expected_headers,
                        verify=ANY,
                    )
                    mock_del_cache.assert_called()
                    assert vault_return.json() == json_success
                    assert mock.call_count == 1


def test_unlimited_use_token_no_decrement(json_success, cache_unlimited):
    """
    Given unlimited-use token, function should succeed not del cache or decrement
    """
    mock = _mock_json_response(json_success)
    expected_headers = {"X-Vault-Token": "test", "Content-Type": "application/json"}
    with patch.object(vault, "get_cache") as mock_get_cache:
        mock_get_cache.return_value = cache_unlimited
        with patch("requests.request", mock):
            with patch.object(vault, "del_cache") as mock_del_cache:
                with patch.object(vault, "write_cache") as mock_write_cache:
                    vault_return = vault.make_request("/secret/my/secret", "key")
                    mock.assert_called_with(
                        "/secret/my/secret",
                        "http://127.0.0.1:8200/key",
                        headers=expected_headers,
                        verify=ANY,
                    )
                    assert (
                        not mock_del_cache.called
                    ), "del cache should not be called for unlimited use token"
                    assert (
                        not mock_write_cache.called
                    ), "write cache should not be called for unlimited use token"
                    assert vault_return.json() == json_success
                    assert mock.call_count == 1


def test_get_cache_standard(cache_single):
    """
    test standard first run of no cache file. Should generate new connection and write cache
    """
    with patch.object(vault, "_read_cache_file") as mock_read_cache:
        mock_read_cache.return_value = {}
        with patch.object(vault, "get_vault_connection") as mock_get_vault_connection:
            mock_get_vault_connection.return_value = copy(cache_single)
            with patch.object(vault, "write_cache") as mock_write_cache:
                cache_result = vault.get_cache()
                mock_write_cache.assert_called_with(copy(cache_single))


def test_get_cache_existing_cache_valid(cache_uses):
    """
    test standard valid cache file
    """
    with patch("time.time", return_value=1234):
        with patch.object(vault, "_read_cache_file") as mock_read_cache:
            mock_read_cache.return_value = cache_uses
            with patch.object(vault, "write_cache") as mock_write_cache:
                with patch.object(vault, "del_cache") as mock_del_cache:
                    cache_result = vault.get_cache()
                    assert not mock_write_cache.called
                    assert not mock_del_cache.called
                    assert cache_result == cache_uses


def test_get_cache_existing_cache_old(cache_uses):
    """
    test old cache file
    """
    with patch("time.time", return_value=3101):
        with patch.object(vault, "get_vault_connection") as mock_get_vault_connection:
            mock_get_vault_connection.return_value = cache_uses
            with patch.object(vault, "_read_cache_file") as mock_read_cache:
                mock_read_cache.return_value = cache_uses
                with patch.object(vault, "write_cache") as mock_write_cache:
                    with patch.object(vault, "del_cache") as mock_del_cache:
                        cache_result = vault.get_cache()
                        assert mock_del_cache.called
                        assert mock_write_cache.called
                        assert cache_result == cache_uses


def test_write_cache_standard(cache_single):
    """
    Test write cache with standard single use token
    """
    function_response = vault.write_cache(copy(cache_single))
    assert vault.__context__["vault_token"] == copy(cache_single)
    assert function_response is True


def test_path_is_v2(metadata_v2):
    """
    Validated v2 path is detected as vault kv v2
    """
    expected_return = {
        "v2": True,
        "data": "secret/data/mything",
        "metadata": "secret/metadata/mything",
        "delete": "secret/mything",
        "type": "kv",
        "destroy": "secret/destroy/mything",
    }
    with patch.object(vault, "_get_secret_path_metadata") as mock_get_metadata:
        mock_get_metadata.return_value = metadata_v2
        function_return = vault.is_v2("secret/mything")
        assert function_return == expected_return


def test_request_with_namespace(json_success, cache_single_namespace):
    """
    Test request with namespace configured
    """
    mock = _mock_json_response(json_success)
    expected_headers = {
        "X-Vault-Token": "test",
        "X-Vault-Namespace": "test_namespace",
        "Content-Type": "application/json",
    }
    supplied_config = {"namespace": "test_namespace"}
    supplied_context = {"vault_token": copy(cache_single_namespace)}
    with patch.dict(vault.__context__, supplied_context):
        with patch.dict(vault.__opts__["vault"], supplied_config):
            with patch("requests.request", mock):
                vault_return = vault.make_request("/secret/my/secret", "key")
                mock.assert_called_with(
                    "/secret/my/secret",
                    "http://127.0.0.1:8200/key",
                    headers=expected_headers,
                    verify=ANY,
                )
                assert vault_return.json() == json_success


def test_get_secret_path_metadata_no_cache(metadata_v2, cache_uses, cache_secret_meta):
    """
    test with no cache file
    """
    make_request_response = {
        "request_id": "b82f2df7-a9b6-920c-0ed2-a3463b996f9e",
        "lease_id": "",
        "renewable": False,
        "lease_duration": 0,
        "data": metadata_v2,
        "wrap_info": None,
        "warnings": None,
        "auth": None,
    }
    cache_object = copy(cache_uses)
    expected_cache_object = copy(cache_uses)
    expected_cache_object.update(copy(cache_secret_meta))
    secret_path = "secret/mything"
    mock = _mock_json_response(make_request_response)
    with patch.object(vault, "_read_cache_file") as mock_read_cache:
        mock_read_cache.return_value = cache_object
        with patch.object(vault, "write_cache") as mock_write_cache:
            with patch("salt.utils.vault.make_request", mock):
                function_result = vault._get_secret_path_metadata(secret_path)
                assert function_result == metadata_v2
                mock_write_cache.assert_called_with(cache_object)
                assert cache_object == expected_cache_object


def test_expand_pattern_lists():
    """
    Ensure expand_pattern_lists works as intended:
    - Expand list-valued patterns
    - Do not change non-list-valued tokens
    """
    cases = {
        "no-tokens-to-replace": ["no-tokens-to-replace"],
        "single-dict:{minion}": ["single-dict:{minion}"],
        "single-list:{grains[roles]}": ["single-list:web", "single-list:database"],
        "multiple-lists:{grains[roles]}+{grains[aux]}": [
            "multiple-lists:web+foo",
            "multiple-lists:web+bar",
            "multiple-lists:database+foo",
            "multiple-lists:database+bar",
        ],
        "single-list-with-dicts:{grains[id]}+{grains[roles]}+{grains[id]}": [
            "single-list-with-dicts:{grains[id]}+web+{grains[id]}",
            "single-list-with-dicts:{grains[id]}+database+{grains[id]}",
        ],
        "deeply-nested-list:{grains[deep][foo][bar][baz]}": [
            "deeply-nested-list:hello",
            "deeply-nested-list:world",
        ],
    }

    pattern_vars = {
        "id": "test-minion",
        "roles": ["web", "database"],
        "aux": ["foo", "bar"],
        "deep": {"foo": {"bar": {"baz": ["hello", "world"]}}},
    }

    mappings = {"minion": "test-minion", "grains": pattern_vars}
    for case, correct_output in cases.items():
        output = vault.expand_pattern_lists(case, **mappings)
        assert output == correct_output


@pytest.mark.parametrize(
    "conf_location,called",
    [("local", False), ("master", True), (None, False), ("doesnotexist", False)],
)
def test_get_vault_connection_config_location(tmp_path, conf_location, called, caplog):
    """
    test the get_vault_connection function when
    config_location is set in opts
    """
    token_url = {
        "url": "http://127.0.0.1",
        "namespace": None,
        "token": "test",
        "verify": None,
        "issued": 1666100373,
        "ttl": 3600,
    }

    opts = {"config_location": conf_location, "pki_dir": tmp_path / "pki"}
    with patch.object(vault, "_get_token_and_url_from_master") as patch_token:
        patch_token.return_vaule = token_url
        with patch.dict(vault.__opts__["vault"], opts):
            vault.get_vault_connection()

    if called:
        patch_token.assert_called()
    else:
        patch_token.assert_not_called()
    if conf_location == "doesnotexist":
        assert "config_location must be either local or master" in caplog.text


def test_del_cache(tmp_cache):
    token_file = tmp_cache / "salt_vault_token"
    token_file.touch()
    with patch.dict(vault.__context__, {"vault_token": "fake_token"}):
        vault.del_cache()
        assert "vault_token" not in vault.__context__
    assert not token_file.exists()



"""
VirtualBox Guest Additions installer state
"""

import logging

log = logging.getLogger(__name__)


def additions_installed(name, reboot=False, upgrade_os=False):
    """
    Ensure that the VirtualBox Guest Additions are installed. Uses the CD,
    connected by VirtualBox.

    name
        The name has no functional value and is only used as a tracking
        reference.
    reboot : False
        Restart OS to complete installation.
    upgrade_os : False
        Upgrade OS (to ensure the latests version of kernel and developer tools
        installed).
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    current_state = __salt__["vbox_guest.additions_version"]()
    if current_state:
        ret["result"] = True
        ret["comment"] = "System already in the correct state"
        return ret
    if __opts__["test"]:
        ret["comment"] = "The state of VirtualBox Guest Additions will be changed."
        ret["changes"] = {
            "old": current_state,
            "new": True,
        }
        ret["result"] = None
        return ret

    new_state = __salt__["vbox_guest.additions_install"](
        reboot=reboot, upgrade_os=upgrade_os
    )

    ret["comment"] = "The state of VirtualBox Guest Additions was changed!"
    ret["changes"] = {
        "old": current_state,
        "new": new_state,
    }
    ret["result"] = bool(new_state)
    return ret


def additions_removed(name, force=False):
    """
    Ensure that the VirtualBox Guest Additions are removed. Uses the CD,
    connected by VirtualBox.

    To connect VirtualBox Guest Additions via VirtualBox graphical interface
    press 'Host+D' ('Host' is usually 'Right Ctrl').

    name
        The name has no functional value and is only used as a tracking
        reference.
    force
        Force VirtualBox Guest Additions removing.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    current_state = __salt__["vbox_guest.additions_version"]()
    if not current_state:
        ret["result"] = True
        ret["comment"] = "System already in the correct state"
        return ret
    if __opts__["test"]:
        ret["comment"] = "The state of VirtualBox Guest Additions will be changed."
        ret["changes"] = {
            "old": current_state,
            "new": True,
        }
        ret["result"] = None
        return ret

    new_state = __salt__["vbox_guest.additions_remove"](force=force)

    ret["comment"] = "The state of VirtualBox Guest Additions was changed!"
    ret["changes"] = {
        "old": current_state,
        "new": new_state,
    }
    ret["result"] = bool(new_state)
    return ret


def grant_access_to_shared_folders_to(name, users=None):
    """
    Grant access to auto-mounted shared folders to the users.

    User is specified by its name. To grant access for several users use
    argument `users`.

    name
        Name of the user to grant access to auto-mounted shared folders to.
    users
        List of names of users to grant access to auto-mounted shared folders to.
        If specified, `name` will not be taken into account.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    current_state = __salt__["vbox_guest.list_shared_folders_users"]()
    if users is None:
        users = [name]
    if current_state == users:
        ret["result"] = True
        ret["comment"] = "System already in the correct state"
        return ret
    if __opts__["test"]:
        ret["comment"] = (
            "List of users who have access to auto-mounted "
            "shared folders will be changed"
        )
        ret["changes"] = {
            "old": current_state,
            "new": users,
        }
        ret["result"] = None
        return ret

    new_state = __salt__["vbox_guest.grant_access_to_shared_folders_to"](
        name=name, users=users
    )

    ret[
        "comment"
    ] = "List of users who have access to auto-mounted shared folders was changed"
    ret["changes"] = {
        "old": current_state,
        "new": new_state,
    }
    ret["result"] = True
    return ret



"""
Management of Microsoft SQLServer Databases
===========================================

The mssql_role module is used to create
and manage SQL Server Roles

.. code-block:: yaml

    yolo:
      mssql_role.present
"""


def __virtual__():
    """
    Only load if the mssql module is present
    """
    if "mssql.version" in __salt__:
        return True
    return (False, "mssql module could not be loaded")


def present(name, owner=None, grants=None, **kwargs):
    """
    Ensure that the named database is present with the specified options

    name
        The name of the database to manage
    owner
        Adds owner using AUTHORIZATION option
    Grants
        Can only be a list of strings
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if __salt__["mssql.role_exists"](name, **kwargs):
        ret[
            "comment"
        ] = "Role {} is already present (Not going to try to set its grants)".format(
            name
        )
        return ret
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Role {} is set to be added".format(name)
        return ret

    role_created = __salt__["mssql.role_create"](
        name, owner=owner, grants=grants, **kwargs
    )
    if (
        role_created is not True
    ):  # Non-empty strings are also evaluated to True, so we cannot use if not role_created:
        ret["result"] = False
        ret["comment"] += "Role {} failed to be created: {}".format(name, role_created)
        return ret
    ret["comment"] += "Role {} has been added".format(name)
    ret["changes"][name] = "Present"
    return ret


def absent(name, **kwargs):
    """
    Ensure that the named database is absent

    name
        The name of the database to remove
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if not __salt__["mssql.role_exists"](name):
        ret["comment"] = "Role {} is not present".format(name)
        return ret
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Role {} is set to be removed".format(name)
        return ret
    if __salt__["mssql.role_remove"](name, **kwargs):
        ret["comment"] = "Role {} has been removed".format(name)
        ret["changes"][name] = "Absent"
        return ret
    # else:
    ret["result"] = False
    ret["comment"] = "Role {} failed to be removed".format(name)
    return ret



def test_deprecation_warning_emits_deprecation_warnings(salt_call_cli):
    ret = salt_call_cli.run("test.deprecation_warning")
    assert ret.stderr.count("DeprecationWarning") >= 2
    assert "This is a test deprecation warning by version." in ret.stderr
    assert (
        "This is a test deprecation warning by date very far into the future (3000-01-01)"
        in ret.stderr
    )



import salt.modules.uwsgi as uwsgi
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, patch
from tests.support.unit import TestCase


class UwsgiTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        patcher = patch("salt.utils.path.which", Mock(return_value="/usr/bin/uwsgi"))
        patcher.start()
        self.addCleanup(patcher.stop)
        return {uwsgi: {}}

    def test_uwsgi_stats(self):
        socket = "127.0.0.1:5050"
        mock = MagicMock(return_value='{"a": 1, "b": 2}')
        with patch.dict(uwsgi.__salt__, {"cmd.run": mock}):
            result = uwsgi.stats(socket)
            mock.assert_called_once_with(
                ["uwsgi", "--connect-and-read", "{}".format(socket)],
                python_shell=False,
            )
            self.assertEqual(result, {"a": 1, "b": 2})



import subprocess

import pytest

from tests.support.helpers import TESTS_DIR


@pytest.mark.parametrize("exp_ret,user_arg", [(1, "false"), (0, "true")])
def test_python_script(install_salt, exp_ret, user_arg):
    ret = subprocess.run(
        install_salt.binary_paths["python"]
        + [str(TESTS_DIR / "files" / "check_python.py"), user_arg],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        universal_newlines=True,
    )

    assert ret.returncode == exp_ret, ret.stderr


def test_python_script_exception(install_salt):
    ret = subprocess.run(
        install_salt.binary_paths["python"]
        + [str(TESTS_DIR / "files" / "check_python.py"), "raise"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        universal_newlines=True,
    )
    assert "Exception: test" in ret.stderr



import platform

import pytest

import salt.utils.files


@pytest.mark.skipif(
    not any([x for x in ["ubuntu", "debian"] if x in platform.platform()]),
    reason="Test only for debian based platforms",
)
def test_adding_repo_file(states, tmp_path):
    """
    test adding a repo file using pkgrepo.managed
    """
    repo_file = str(tmp_path / "stable-binary.list")
    repo_content = "deb http://www.deb-multimedia.org stable main"
    ret = states.pkgrepo.managed(name=repo_content, file=repo_file, clean_file=True)
    with salt.utils.files.fopen(repo_file, "r") as fp:
        file_content = fp.read()
    assert file_content.strip() == repo_content


@pytest.mark.skipif(
    not any([x for x in ["ubuntu", "debian"] if x in platform.platform()]),
    reason="Test only for debian based platforms",
)
def test_adding_repo_file_arch(states, tmp_path):
    """
    test adding a repo file using pkgrepo.managed
    and setting architecture
    """
    repo_file = str(tmp_path / "stable-binary.list")
    repo_content = "deb [arch=amd64  ] http://www.deb-multimedia.org stable main"
    ret = states.pkgrepo.managed(name=repo_content, file=repo_file, clean_file=True)
    with salt.utils.files.fopen(repo_file, "r") as fp:
        file_content = fp.read()
        assert (
            file_content.strip()
            == "deb [arch=amd64] http://www.deb-multimedia.org stable main"
        )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.boto_ec2 as boto_ec2
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {boto_ec2: {}}


def test_key_present():
    """
    Test to ensure key pair is present.
    """
    name = "mykeypair"
    upublic = "salt://mybase/public_key.pub"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[True, False, False])
    mock_bool = MagicMock(side_effect=[IOError, True])
    with patch.dict(
        boto_ec2.__salt__, {"boto_ec2.get_key": mock, "cp.get_file_str": mock_bool}
    ):
        comt = "The key name {} already exists".format(name)
        ret.update({"comment": comt})
        assert boto_ec2.key_present(name) == ret

        comt = "File {} not found.".format(upublic)
        ret.update({"comment": comt, "result": False})
        assert boto_ec2.key_present(name, upload_public=upublic) == ret

        with patch.dict(boto_ec2.__opts__, {"test": True}):
            comt = "The key {} is set to be created.".format(name)
            ret.update({"comment": comt, "result": None})
            assert boto_ec2.key_present(name, upload_public=upublic) == ret


def test_key_absent():
    """
    Test to delete a key pair
    """
    name = "new_table"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[False, True])
    with patch.dict(boto_ec2.__salt__, {"boto_ec2.get_key": mock}):
        comt = "The key name {} does not exist".format(name)
        ret.update({"comment": comt})
        assert boto_ec2.key_absent(name) == ret

        with patch.dict(boto_ec2.__opts__, {"test": True}):
            comt = "The key {} is set to be deleted.".format(name)
            ret.update({"comment": comt, "result": None})
            assert boto_ec2.key_absent(name) == ret



"""
Manage SNS Topics
=================

Create and destroy SNS topics. Be aware that this interacts with Amazon's
services, and so may incur charges.

This module uses ``boto``, which can be installed via package, or pip.

This module accepts explicit AWS credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More information available `here
<http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    sns.keyid: GKTADJGHEIQSXMKKRBJ08H
    sns.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify ``key``, ``keyid`` and ``region`` via a profile, either
passed in as a dict, or as a string to pull from pillars or minion config:

.. code-block:: yaml

    myprofile:
        keyid: GKTADJGHEIQSXMKKRBJ08H
        key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
        region: us-east-1

.. code-block:: yaml

    mytopic:
        boto3_sns.topic_present:
            - region: us-east-1
            - keyid: GKTADJGHEIQSXMKKRBJ08H
            - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

Using a profile from pillars
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. code-block:: yaml

    mytopic:
        boto3_sns.topic_present:
            - region: us-east-1
            - profile: mysnsprofile

Passing in a profile
^^^^^^^^^^^^^^^^^^^^

.. code-block:: yaml

    mytopic:
        boto3_sns.topic_present:
            - region: us-east-1
            - profile:
                keyid: GKTADJGHEIQSXMKKRBJ08H
                key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
"""

import copy
import logging
import re

import salt.utils.json

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto is available.
    """
    if "boto3_sns.topic_exists" in __salt__:
        return "boto3_sns"
    return (False, "boto3_sns module could not be loaded")


def topic_present(
    name,
    subscriptions=None,
    attributes=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure the SNS topic exists.

    name
        Name of the SNS topic.

    subscriptions
        List of SNS subscriptions.

        Each subscription is a dictionary with a protocol and endpoint key:

        .. code-block:: yaml

            subscriptions:
            - Protocol: https
              Endpoint: https://www.example.com/sns-endpoint
            - Protocol: sqs
              Endpoint: arn:aws:sqs:us-west-2:123456789012:MyQueue

    attributes
        Dictionary of attributes to set on the SNS topic
        Valid attribute keys are:

          - Policy:  the JSON serialization of the topic's access control policy
          - DisplayName:  the human-readable name used in the "From" field for notifications
                to email and email-json endpoints
          - DeliveryPolicy:  the JSON serialization of the topic's delivery policy

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    something_changed = False
    current = __salt__["boto3_sns.describe_topic"](name, region, key, keyid, profile)
    if current:
        ret["comment"] = "AWS SNS topic {} present.".format(name)
        TopicArn = current["TopicArn"]
    else:
        if __opts__["test"]:
            ret["comment"] = "AWS SNS topic {} would be created.".format(name)
            ret["result"] = None
            return ret
        else:
            TopicArn = __salt__["boto3_sns.create_topic"](
                name, region=region, key=key, keyid=keyid, profile=profile
            )
            if TopicArn:
                ret["comment"] = "AWS SNS topic {} created with ARN {}.".format(
                    name, TopicArn
                )
                something_changed = True
            else:
                ret["comment"] = "Failed to create AWS SNS topic {}".format(name)
                log.error(ret["comment"])
                ret["result"] = False
                return ret

    ### Update any explicitly defined attributes
    want_attrs = attributes if attributes else {}
    # Freshen these in case we just created it above
    current_attrs = __salt__["boto3_sns.get_topic_attributes"](
        TopicArn, region=region, key=key, keyid=keyid, profile=profile
    )
    for attr in ["DisplayName", "Policy", "DeliveryPolicy"]:
        curr_val = current_attrs.get(attr)
        want_val = want_attrs.get(attr)
        # Some get default values if not set, so it's not safe to enforce absence if they're
        # not provided at all.  This implies that if you want to clear a value, you must explicitly
        # set it to an empty string.
        if want_val is None:
            continue
        if _json_objs_equal(want_val, curr_val):
            continue
        if __opts__["test"]:
            ret["comment"] += "  Attribute {} would be updated on topic {}.".format(
                attr, TopicArn
            )
            ret["result"] = None
            continue
        want_val = (
            want_val if isinstance(want_val, str) else salt.utils.json.dumps(want_val)
        )
        if __salt__["boto3_sns.set_topic_attributes"](
            TopicArn,
            attr,
            want_val,
            region=region,
            key=key,
            keyid=keyid,
            profile=profile,
        ):
            ret["comment"] += "  Attribute {} set to {} on topic {}.".format(
                attr, want_val, TopicArn
            )
            something_changed = True
        else:
            ret["comment"] += "  Failed to update {} on topic {}.".format(
                attr, TopicArn
            )
            ret["result"] = False
            return ret

    ### Add / remove subscriptions
    want_subs = subscriptions if subscriptions else []
    obfuscated_subs = []
    current_subs = current.get("Subscriptions", [])
    current_slim = [
        {"Protocol": s["Protocol"], "Endpoint": s["Endpoint"]} for s in current_subs
    ]
    subscribe = []
    unsubscribe = []
    for sub in want_subs:
        # If the subscription contains inline digest auth, AWS will obfuscate the password with
        # '****'.  Thus we need to do the same with ours to permit 1-to-1 comparison.
        # Example: https://user:****@my.endpoiint.com/foo/bar
        endpoint = sub["Endpoint"]
        matches = re.search(r"https://(?P<user>\w+):(?P<pass>\w+)@", endpoint)
        if matches is not None:
            sub["Endpoint"] = endpoint.replace(
                ":" + matches.groupdict()["pass"], ":****"
            )
        obfuscated_subs += [copy.deepcopy(sub)]
        # Now set it back...
        if sub not in current_slim:
            sub["Endpoint"] = endpoint
            subscribe += [sub]
    for sub in current_subs:
        minimal = {"Protocol": sub["Protocol"], "Endpoint": sub["Endpoint"]}
        if minimal not in obfuscated_subs and sub["SubscriptionArn"].startswith(
            "arn:aws:sns:"
        ):
            unsubscribe += [sub["SubscriptionArn"]]
    for sub in subscribe:
        prot = sub["Protocol"]
        endp = sub["Endpoint"]
        if __opts__["test"]:
            msg = " Subscription {}:{} would be set on topic {}.".format(
                prot, endp, TopicArn
            )
            ret["comment"] += msg
            ret["result"] = None
            continue
        subbed = __salt__["boto3_sns.subscribe"](
            TopicArn, prot, endp, region=region, key=key, keyid=keyid, profile=profile
        )
        if subbed:
            msg = " Subscription {}:{} set on topic {}.".format(prot, endp, TopicArn)
            ret["comment"] += msg
            something_changed = True
        else:
            msg = " Failed to set subscription {}:{} on topic {}.".format(
                prot, endp, TopicArn
            )
            ret["comment"] += msg
            ret["result"] = False
            return ret
    for sub in unsubscribe:
        if __opts__["test"]:
            msg = "  Subscription {} would be removed from topic {}.".format(
                sub, TopicArn
            )
            ret["comment"] += msg
            ret["result"] = None
            continue
        unsubbed = __salt__["boto3_sns.unsubscribe"](
            sub, region=region, key=key, keyid=keyid, profile=profile
        )
        if unsubbed:
            ret["comment"] += "  Subscription {} removed from topic {}.".format(
                sub, TopicArn
            )
            something_changed = True
        else:
            msg = "  Failed to remove subscription {} from topic {}.".format(
                sub, TopicArn
            )
            ret["comment"] += msg
            ret["result"] = False
            return ret
    if something_changed:
        ret["changes"]["old"] = current
        ret["changes"]["new"] = __salt__["boto3_sns.describe_topic"](
            name, region, key, keyid, profile
        )
    return ret


def topic_absent(
    name, unsubscribe=False, region=None, key=None, keyid=None, profile=None
):
    """
    Ensure the named sns topic is deleted.

    name
        Name of the SNS topic.

    unsubscribe
        If True, unsubscribe all subcriptions to the SNS topic before
        deleting the SNS topic

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string)
        that contains a dict with region, key and keyid.
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    something_changed = False
    current = __salt__["boto3_sns.describe_topic"](name, region, key, keyid, profile)
    if not current:
        ret["comment"] = "AWS SNS topic {} absent.".format(name)
    else:
        TopicArn = current["TopicArn"]
        if __opts__["test"]:
            ret["comment"] = "AWS SNS topic {} would be removed.".format(TopicArn)
            if unsubscribe:
                ret["comment"] += "  {} subscription(s) would be removed.".format(
                    len(current["Subscriptions"])
                )
            ret["result"] = None
            return ret
        if unsubscribe:
            for sub in current["Subscriptions"]:
                if sub["SubscriptionArn"] == "PendingConfirmation":
                    # The API won't let you delete subscriptions in pending status...
                    log.warning(
                        "Ignoring PendingConfirmation subscription %s %s on topic %s",
                        sub["Protocol"],
                        sub["Endpoint"],
                        sub["TopicArn"],
                    )
                    continue
                if __salt__["boto3_sns.unsubscribe"](
                    sub["SubscriptionArn"],
                    region=region,
                    key=key,
                    keyid=keyid,
                    profile=profile,
                ):
                    log.debug("Deleted subscription %s for SNS topic %s", sub, TopicArn)
                    something_changed = True
                else:
                    ret[
                        "comment"
                    ] = "Failed to delete subscription {} for SNS topic {}".format(
                        sub, TopicArn
                    )
                    ret["result"] = False
                    return ret
        if not __salt__["boto3_sns.delete_topic"](
            TopicArn, region=region, key=key, keyid=keyid, profile=profile
        ):
            ret["comment"] = "Failed to delete SNS topic {}".format(TopicArn)
            log.error(ret["comment"])
            ret["result"] = False
        else:
            ret["comment"] = "AWS SNS topic {} deleted.".format(TopicArn)
            if unsubscribe:
                ret["comment"] += "  ".join(
                    [
                        "Subscription {} deleted".format(s)
                        for s in current["Subscriptions"]
                    ]
                )
            something_changed = True

    if something_changed:
        ret["changes"]["old"] = current
        ret["changes"]["new"] = __salt__["boto3_sns.describe_topic"](
            name, region, key, keyid, profile
        )
    return ret


def _json_objs_equal(left, right):
    left = __utils__["boto3.ordered"](
        salt.utils.json.loads(left) if isinstance(left, str) else left
    )
    right = __utils__["boto3.ordered"](
        salt.utils.json.loads(right) if isinstance(right, str) else right
    )
    return left == right



import os

import pytest

from tests.support.mock import patch
from tests.support.unit import TestCase

try:
    import salt.utils.win_system as win_system
except Exception as exc:  # pylint: disable=broad-except
    win_system = exc


class WinSystemImportTestCase(TestCase):
    """
    Simply importing should not raise an error, especially on Linux
    """

    def test_import(self):
        if isinstance(win_system, Exception):
            raise Exception(
                "Importing win_system caused traceback: {}".format(win_system)
            )


@pytest.mark.skip_unless_on_windows
class WinSystemTestCase(TestCase):
    """
    Test cases for salt.utils.win_system
    """

    def test_get_computer_name(self):
        """
        Should return the computer name
        """
        with patch("win32api.GetComputerNameEx", return_value="FAKENAME"):
            self.assertEqual(win_system.get_computer_name(), "FAKENAME")

    def test_get_computer_name_fail(self):
        """
        If it fails, it returns False
        """
        with patch("win32api.GetComputerNameEx", return_value=None):
            self.assertFalse(win_system.get_computer_name())

    def test_get_pending_computer_name(self):
        """
        Will return the pending computer name if one is pending
        """
        expected = "PendingName"
        patch_value = {"vdata": expected}
        with patch("salt.utils.win_reg.read_value", return_value=patch_value):
            result = win_system.get_pending_computer_name()
            self.assertEqual(expected, result)

    def test_get_pending_computer_name_none(self):
        """
        Will return the None if the pending computer is the current name
        """
        patch_value = {"vdata": os.environ.get("COMPUTERNAME")}
        with patch("salt.utils.win_reg.read_value", return_value=patch_value):
            self.assertIsNone(win_system.get_pending_computer_name())

    def test_get_pending_computer_name_false(self):
        """
        Will return False if there is no pending computer name
        """
        with patch("salt.utils.win_reg.read_value", return_value=False):
            self.assertIsNone(win_system.get_pending_computer_name())

    def test_get_pending_component_servicing(self):
        """
        If none of the keys exist, should return False
        """
        with patch("salt.utils.win_reg.key_exists", return_value=False):
            self.assertFalse(win_system.get_pending_component_servicing())

    def test_get_pending_component_servicing_true_1(self):
        """
        If the RebootPending key exists, should return True
        """
        with patch("salt.utils.win_reg.key_exists", side_effect=[True]):
            self.assertTrue(win_system.get_pending_component_servicing())

    def test_get_pending_component_servicing_true_2(self):
        """
        If the RebootInProgress key exists, should return True
        """
        with patch("salt.utils.win_reg.key_exists", side_effect=[False, True]):
            self.assertTrue(win_system.get_pending_component_servicing())

    def test_get_pending_component_servicing_true_3(self):
        """
        If the PackagesPending key exists, should return True
        """
        with patch("salt.utils.win_reg.key_exists", side_effect=[False, False, True]):
            self.assertTrue(win_system.get_pending_component_servicing())

    def test_get_pending_domain_join(self):
        """
        If none of the keys exist, should return False
        """
        with patch("salt.utils.win_reg.key_exists", return_value=False):
            self.assertFalse(win_system.get_pending_domain_join())

    def test_get_pending_domain_join_true_1(self):
        """
        If the AvoidSpnSet key exists, should return True
        """
        with patch("salt.utils.win_reg.key_exists", side_effect=[True]):
            self.assertTrue(win_system.get_pending_domain_join())

    def test_get_pending_domain_join_true_2(self):
        """
        If the JoinDomain key exists, should return True
        """
        with patch("salt.utils.win_reg.key_exists", side_effect=[False, True]):
            self.assertTrue(win_system.get_pending_domain_join())

    def test_get_pending_file_rename_false_1(self):
        """
        If none of the value names exist, should return False
        """
        patched_return = {"success": False}
        with patch("salt.utils.win_reg.read_value", return_value=patched_return):
            self.assertFalse(win_system.get_pending_file_rename())

    def test_get_pending_file_rename_false_2(self):
        """
        If one of the value names exists but is not set, should return False
        """
        patched_return = {"success": True, "vdata": "(value not set)"}
        with patch("salt.utils.win_reg.read_value", return_value=patched_return):
            self.assertFalse(win_system.get_pending_file_rename())

    def test_get_pending_file_rename_true_1(self):
        """
        If one of the value names exists and is set, should return True
        """
        patched_return = {"success": True, "vdata": "some value"}
        with patch("salt.utils.win_reg.read_value", return_value=patched_return):
            self.assertTrue(win_system.get_pending_file_rename())

    def test_get_pending_servermanager_false_1(self):
        """
        If the CurrentRebootAttempts value name does not exist, should return
        False
        """
        patched_return = {"success": False}
        with patch("salt.utils.win_reg.read_value", return_value=patched_return):
            self.assertFalse(win_system.get_pending_servermanager())

    def test_get_pending_servermanager_false_2(self):
        """
        If the CurrentRebootAttempts value name exists but is not an integer,
        should return False
        """
        patched_return = {"success": True, "vdata": "(value not set)"}
        with patch("salt.utils.win_reg.read_value", return_value=patched_return):
            self.assertFalse(win_system.get_pending_file_rename())

    def test_get_pending_servermanager_true(self):
        """
        If the CurrentRebootAttempts value name exists and is an integer,
        should return True
        """
        patched_return = {"success": True, "vdata": 1}
        with patch("salt.utils.win_reg.read_value", return_value=patched_return):
            self.assertTrue(win_system.get_pending_file_rename())

    def test_get_pending_dvd_reboot(self):
        """
        If the DVDRebootSignal value name does not exist, should return False
        """
        with patch("salt.utils.win_reg.value_exists", return_value=False):
            self.assertFalse(win_system.get_pending_dvd_reboot())

    def test_get_pending_dvd_reboot_true(self):
        """
        If the DVDRebootSignal value name exists, should return True
        """
        with patch("salt.utils.win_reg.value_exists", return_value=True):
            self.assertTrue(win_system.get_pending_dvd_reboot())

    def test_get_pending_update(self):
        """
        If none of the keys exist and there are not subkeys, should return False
        """
        with patch("salt.utils.win_reg.key_exists", return_value=False), patch(
            "salt.utils.win_reg.list_keys", return_value=[]
        ):
            self.assertFalse(win_system.get_pending_update())

    def test_get_pending_update_true_1(self):
        """
        If the RebootRequired key exists, should return True
        """
        with patch("salt.utils.win_reg.key_exists", side_effect=[True]):
            self.assertTrue(win_system.get_pending_update())

    def test_get_pending_update_true_2(self):
        """
        If the PostRebootReporting key exists, should return True
        """
        with patch("salt.utils.win_reg.key_exists", side_effect=[False, True]):
            self.assertTrue(win_system.get_pending_update())

    def test_get_reboot_required_witnessed_false_1(self):
        """
        The ``Reboot Required`` value name does not exist, should return False
        """
        patched_data = {"vdata": None}
        with patch("salt.utils.win_reg.read_value", return_value=patched_data):
            self.assertFalse(win_system.get_reboot_required_witnessed())

    def test_get_reboot_required_witnessed_false_2(self):
        """
        The ``Reboot required`` value name is set to 0, should return False
        """
        patched_data = {"vdata": 0}
        with patch("salt.utils.win_reg.read_value", return_value=patched_data):
            self.assertFalse(win_system.get_reboot_required_witnessed())

    def test_get_reboot_required_witnessed_true(self):
        """
        The ``Reboot required`` value name is set to 1, should return True
        """
        patched_data = {"vdata": 1}
        with patch("salt.utils.win_reg.read_value", return_value=patched_data):
            self.assertTrue(win_system.get_reboot_required_witnessed())

    def test_set_reboot_required_witnessed(self):
        """
        The call to ``set_value`` should return True and should be called with
        the specified parameters
        """
        with patch("salt.utils.win_reg.set_value", return_value=True) as sv:
            self.assertTrue(win_system.set_reboot_required_witnessed())
            sv.assert_called_once_with(
                hive="HKLM",
                key=win_system.MINION_VOLATILE_KEY,
                volatile=True,
                vname=win_system.REBOOT_REQUIRED_NAME,
                vdata=1,
                vtype="REG_DWORD",
            )

    def test_get_pending_update_exe_volatile_false_1(self):
        """
        If UpdateExeVolatile value name is 0, should return False
        """
        patched_data = {"success": True, "vdata": 0}
        with patch("salt.utils.win_reg.read_value", return_value=patched_data):
            self.assertFalse(win_system.get_pending_update_exe_volatile())

    def test_get_pending_update_exe_volatile_false_2(self):
        """
        If UpdateExeVolatile value name is not present, should return False
        """
        patched_data = {"success": False}
        with patch("salt.utils.win_reg.read_value", return_value=patched_data):
            self.assertFalse(win_system.get_pending_update_exe_volatile())

    def test_get_pending_update_exe_volatile_true_1(self):
        """
        If UpdateExeVolatile value name is not 0, should return True
        """
        patched_data = {"success": True, "vdata": 1}
        with patch("salt.utils.win_reg.read_value", return_value=patched_data):
            self.assertTrue(win_system.get_pending_update_exe_volatile())

    def test_get_pending_reboot(self):
        """
        If all functions return Falsy data, should return False
        """
        with patch(
            "salt.utils.win_system.get_pending_update", return_value=False
        ), patch("salt.utils.win_update.needs_reboot", return_value=False), patch(
            "salt.utils.win_system.get_pending_update_exe_volatile", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_file_rename", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_servermanager", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_component_servicing", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_dvd_reboot", return_value=False
        ), patch(
            "salt.utils.win_system.get_reboot_required_witnessed", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_computer_name", return_value=None
        ), patch(
            "salt.utils.win_system.get_pending_domain_join", return_value=False
        ):
            self.assertFalse(win_system.get_pending_reboot())

    def test_get_pending_reboot_true_1(self):
        """
        If any boolean returning functions return True, should return True
        """
        with patch(
            "salt.utils.win_system.get_pending_update", return_value=False
        ), patch("salt.utils.win_update.needs_reboot", return_value=False), patch(
            "salt.utils.win_system.get_pending_update_exe_volatile", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_file_rename", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_servermanager", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_component_servicing", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_dvd_reboot", return_value=False
        ), patch(
            "salt.utils.win_system.get_reboot_required_witnessed", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_computer_name", return_value=None
        ), patch(
            "salt.utils.win_system.get_pending_domain_join", return_value=True
        ):
            self.assertTrue(win_system.get_pending_reboot())

    def test_get_pending_reboot_true_2(self):
        """
        If a computer name is returned, should return True
        """
        with patch(
            "salt.utils.win_system.get_pending_update", return_value=False
        ), patch("salt.utils.win_update.needs_reboot", return_value=False), patch(
            "salt.utils.win_system.get_pending_update_exe_volatile", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_file_rename", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_servermanager", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_component_servicing", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_dvd_reboot", return_value=False
        ), patch(
            "salt.utils.win_system.get_reboot_required_witnessed", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_computer_name",
            return_value="pending name",
        ):
            self.assertTrue(win_system.get_pending_reboot())

    def test_get_pending_reboot_details(self):
        """
        All items False should return a dictionary with all items False
        """
        with patch(
            "salt.utils.win_system.get_pending_update", return_value=False
        ), patch("salt.utils.win_update.needs_reboot", return_value=False), patch(
            "salt.utils.win_system.get_pending_update_exe_volatile", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_file_rename", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_servermanager", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_component_servicing", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_dvd_reboot", return_value=False
        ), patch(
            "salt.utils.win_system.get_reboot_required_witnessed", return_value=False
        ), patch(
            "salt.utils.win_system.get_pending_computer_name", return_value=None
        ), patch(
            "salt.utils.win_system.get_pending_domain_join", return_value=False
        ):
            expected = {
                "Pending Component Servicing": False,
                "Pending Computer Rename": False,
                "Pending DVD Reboot": False,
                "Pending File Rename": False,
                "Pending Join Domain": False,
                "Pending ServerManager": False,
                "Pending Update": False,
                "Pending Windows Update": False,
                "Reboot Required Witnessed": False,
                "Volatile Update Exe": False,
            }
            result = win_system.get_pending_reboot_details()
            self.assertDictEqual(expected, result)

    def test_get_pending_reboot_details_true(self):
        """
        All items True should return a dictionary with all items True
        """
        with patch(
            "salt.utils.win_system.get_pending_update", return_value=True
        ), patch("salt.utils.win_update.needs_reboot", return_value=True), patch(
            "salt.utils.win_system.get_pending_update_exe_volatile", return_value=True
        ), patch(
            "salt.utils.win_system.get_pending_file_rename", return_value=True
        ), patch(
            "salt.utils.win_system.get_pending_servermanager", return_value=True
        ), patch(
            "salt.utils.win_system.get_pending_component_servicing", return_value=True
        ), patch(
            "salt.utils.win_system.get_pending_dvd_reboot", return_value=True
        ), patch(
            "salt.utils.win_system.get_reboot_required_witnessed", return_value=True
        ), patch(
            "salt.utils.win_system.get_pending_computer_name",
            return_value="pending name",
        ), patch(
            "salt.utils.win_system.get_pending_domain_join", return_value=True
        ):
            expected = {
                "Pending Component Servicing": True,
                "Pending Computer Rename": True,
                "Pending DVD Reboot": True,
                "Pending File Rename": True,
                "Pending Join Domain": True,
                "Pending ServerManager": True,
                "Pending Update": True,
                "Pending Windows Update": True,
                "Reboot Required Witnessed": True,
                "Volatile Update Exe": True,
            }
            result = win_system.get_pending_reboot_details()
            self.assertDictEqual(expected, result)



"""
    :codeauthor: :email:`Jorge Schrauwen <sjorge@blackdot.be>`
"""

import pytest

import salt.grains.mdata as mdata
from tests.support.mock import Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        mdata: {"__salt__": {}},
    }


def test_user_mdata_missing_cmd_both():
    """
    When both or either of the commands is missing there should
    be no grain output.
    """
    grains_exp_res = {}

    which_mock = Mock(side_effect=[None, None])
    with patch("salt.utils.path.which", which_mock):
        grains_res = mdata._user_mdata()
        assert grains_exp_res == grains_res


def test_user_mdata_missing_cmd_one():
    """
    When both or either of the commands is missing there should
    be no grain output.
    """
    grains_exp_res = {}

    which_mock = Mock(side_effect=["/usr/sbin/mdata-list", None])
    with patch("salt.utils.path.which", which_mock):
        grains_res = mdata._user_mdata()
        assert grains_exp_res == grains_res


def test_user_mdata_empty_list():
    """
    When there are no user grains, there are no mdata-get calls
    so there are also no grains.
    """
    grains_exp_res = {}

    which_mock = Mock(side_effect=["/usr/sbin/mdata-list", "/usr/sbin/mdata-get"])
    cmd_mock = Mock(side_effect=[""])
    with patch("salt.utils.path.which", which_mock), patch.dict(
        mdata.__salt__, {"cmd.run": cmd_mock}
    ):
        grains_res = mdata._user_mdata()
        assert grains_exp_res == grains_res


def test_user_mdata():
    """
    We have a list of two grains, so there should be two mdata-get
    calls, resulting in 2 grains.
    """
    grains_exp_res = {
        "mdata": {
            "multi_text_data": "multi\nline\ntext",
            "simple_text_data": "some text data",
        },
    }

    which_mock = Mock(side_effect=["/usr/sbin/mdata-list", "/usr/sbin/mdata-get"])
    cmd_mock = Mock(
        side_effect=[
            "simple_text_data\nmulti_text_data",
            "some text data",
            "multi\nline\ntext",
        ]
    )
    with patch("salt.utils.path.which", which_mock), patch.dict(
        mdata.__salt__, {"cmd.run": cmd_mock}
    ):
        grains_res = mdata._user_mdata()

        assert grains_exp_res == grains_res


def test_sdc_mdata_missing_cmd_both():
    """
    When both or either of the commands is missing there should
    be no grain output.
    """
    which_mock = Mock(side_effect=[None, None])
    with patch("salt.utils.path.which", which_mock):
        grains = mdata._sdc_mdata()
        assert grains == {}


def test_sdc_mdata_missing_cmd_one():
    """
    When both or either of the commands is missing there should
    be no grain output.
    """
    grains_exp_res = {}

    which_mock = Mock(side_effect=["/usr/sbin/mdata-list", None])
    with patch("salt.utils.path.which", which_mock):
        grains_res = mdata._sdc_mdata()
        assert grains_exp_res == grains_res


def test_sdc_mdata():
    """
    Simulate all mdata_get calls from a test zone.
    """
    grains_exp_res = {
        "mdata": {
            "sdc": {
                "alias": "test",
                "dns_domain": "example.org",
                "hostname": "test_salt",
                "nics": [
                    {
                        "gateway": "10.12.3.1",
                        "gateways": ["10.12.3.1"],
                        "interface": "net0",
                        "ip": "10.12.3.123",
                        "ips": ["10.12.3.123/24", "2001:ffff:ffff:123::123/64"],
                        "mac": "00:00:00:00:00:01",
                        "mtu": 1500,
                        "netmask": "255.255.255.0",
                        "nic_tag": "trunk",
                        "primary": True,
                        "vlan_id": 123,
                    }
                ],
                "resolvers": ["10.12.3.1", "2001:ffff:ffff:123::1"],
                "routes": [],
                "server_uuid": "00000000-0000-0000-0000-000123456789",
                "uuid": "bae504b1-4594-47de-e2ed-e4f454776689",
            },
        },
    }

    which_mock = Mock(side_effect=["/usr/sbin/mdata-list", "/usr/sbin/mdata-get"])
    cmd_mock = Mock(
        side_effect=[
            "bae504b1-4594-47de-e2ed-e4f454776689",
            "00000000-0000-0000-0000-000123456789",
            "No metadata for 'sdc:datacenter_name'",
            "test_salt",
            "example.org",
            "test",
            '["10.12.3.1","2001:ffff:ffff:123::1"]',
            '[{"interface":"net0","mac":"00:00:00:00:00:01","vlan_id":123,"nic_tag":"trunk","gateway":"10.12.3.1","gateways":["10.12.3.1"],"netmask":"255.255.255.0","ip":"10.12.3.123","ips":["10.12.3.123/24","2001:ffff:ffff:123::123/64"],"mtu":1500,"primary":true}]',
            "[]",
        ]
    )
    with patch("salt.utils.path.which", which_mock), patch.dict(
        mdata.__salt__, {"cmd.run": cmd_mock}
    ):
        grains_res = mdata._sdc_mdata()

        assert grains_exp_res == grains_res



import pytest

import salt.fileserver.s3fs as s3fs


@pytest.fixture
def configure_loader_modules(tmp_path):
    opts = {
        "cachedir": tmp_path,
    }
    return {s3fs: {"__opts__": opts}}


def test_cache_round_trip():
    metadata = {"foo": "bar"}
    cache_file = s3fs._get_cached_file_name("base", "fake_bucket", "some_file")

    s3fs._write_buckets_cache_file(metadata, cache_file)
    assert s3fs._read_buckets_cache_file(cache_file) == metadata


def test_ignore_pickle_load_exceptions():
    #  TODO: parameterized test with patched pickle.load that raises the
    #  various allowable exception from _read_buckets_cache_file
    pass



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.win_shadow
"""


import pytest

import salt.modules.win_shadow as win_shadow
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_shadow: {"__salt__": {"user.update": MagicMock(return_value=True)}}}


def test_info():
    """
    Test if it return information for the specified user
    """
    mock_user_info = MagicMock(
        return_value={"name": "SALT", "password_changed": "", "expiration_date": ""}
    )
    with patch.dict(win_shadow.__salt__, {"user.info": mock_user_info}):
        assert win_shadow.info("SALT") == {
            "name": "SALT",
            "passwd": "Unavailable",
            "lstchg": "",
            "min": "",
            "max": "",
            "warn": "",
            "inact": "",
            "expire": "",
        }


def test_set_password():
    """
    Test if it set the password for a named user.
    """
    mock_cmd = MagicMock(return_value={"retcode": False})
    mock_user_info = MagicMock(
        return_value={"name": "SALT", "password_changed": "", "expiration_date": ""}
    )
    with patch.dict(
        win_shadow.__salt__, {"cmd.run_all": mock_cmd, "user.info": mock_user_info}
    ):
        assert win_shadow.set_password("root", "mysecretpassword")



"""
    :codeauthor: `Anthony Shaw <anthonyshaw@apache.org>`

    tests.unit.cloud.clouds.dimensiondata_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import pytest

from salt.cloud.clouds import dimensiondata
from salt.exceptions import SaltCloudSystemExit
from salt.utils.versions import Version
from tests.support.mock import MagicMock
from tests.support.mock import __version__ as mock_version
from tests.support.mock import patch

try:
    import libcloud.security

    HAS_LIBCLOUD = True
except ImportError:
    HAS_LIBCLOUD = False

# Use certifi if installed
try:
    if HAS_LIBCLOUD:
        # This work-around for Issue #32743 is no longer needed for libcloud >=
        # 1.4.0. However, older versions of libcloud must still be supported
        # with this work-around. This work-around can be removed when the
        # required minimum version of libcloud is 2.0.0 (See PR #40837 - which
        # is implemented in Salt 2018.3.0).
        if Version(libcloud.__version__) < Version("1.4.0"):
            import certifi

            libcloud.security.CA_CERTS_PATH.append(certifi.where())
except (ImportError, NameError):
    pass


@pytest.fixture
def vm_name():
    return "winterfell"


def _preferred_ip(ip_set, preferred=None):
    """
    Returns a function that reacts which ip is preferred
    :param ip_set:
    :param private:
    :return:
    """

    def _ip_decider(vm, ips):
        for ip in ips:
            if ip in preferred:
                return ip
        return False

    return _ip_decider


@pytest.fixture
def configure_loader_modules():
    return {
        dimensiondata: {
            "__active_provider_name__": "",
            "__opts__": {
                "providers": {
                    "my-dimensiondata-cloud": {
                        "dimensiondata": {
                            "driver": "dimensiondata",
                            "region": "dd-au",
                            "user_id": "jon_snow",
                            "key": "IKnowNothing",
                        }
                    }
                }
            },
        }
    }


def test_avail_images_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call avail_images
    with --action or -a.
    """
    with pytest.raises(SaltCloudSystemExit):
        dimensiondata.avail_images(call="action")


def test_avail_locations_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call avail_locations
    with --action or -a.
    """
    with pytest.raises(SaltCloudSystemExit):
        dimensiondata.avail_locations(call="action")


def test_avail_sizes_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call avail_sizes
    with --action or -a.
    """
    with pytest.raises(SaltCloudSystemExit):
        dimensiondata.avail_sizes(call="action")


def test_list_nodes_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call list_nodes
    with --action or -a.
    """
    with pytest.raises(SaltCloudSystemExit):
        dimensiondata.list_nodes(call="action")


def test_destroy_call(vm_name):
    """
    Tests that a SaltCloudSystemExit is raised when trying to call destroy
    with --function or -f.
    """
    with pytest.raises(SaltCloudSystemExit):
        dimensiondata.destroy(name=vm_name, call="function")


@pytest.mark.skipif(
    HAS_LIBCLOUD is False, reason="Install 'libcloud' to be able to run this unit test."
)
def test_avail_sizes():
    """
    Tests that avail_sizes returns an empty dictionary.
    """
    sizes = dimensiondata.avail_sizes(call="foo")
    assert len(sizes) == 1
    assert sizes["default"]["name"] == "default"


def test_import():
    """
    Test that the module picks up installed deps
    """
    with patch("salt.config.check_driver_dependencies", return_value=True) as p:
        get_deps = dimensiondata.get_dependencies()
        assert get_deps is True
        if Version(mock_version) >= Version("2.0.0"):
            assert p.call_count >= 1


def test_provider_matches():
    """
    Test that the first configured instance of a dimensiondata driver is matched
    """
    p = dimensiondata.get_configured_provider()
    assert p is not None


def test_query_node_data_filter_preferred_ip_addresses():
    """
    Test if query node data is filtering out unpreferred IP addresses.
    """
    zero_ip = "0.0.0.0"
    private_ips = [zero_ip, "1.1.1.1", "2.2.2.2"]
    vm = {"name": None}
    data = MagicMock()
    data.public_ips = []
    # pylint: disable=blacklisted-unmocked-patching
    dimensiondata.NodeState = MagicMock()
    # pylint: enable=blacklisted-unmocked-patching
    dimensiondata.NodeState.RUNNING = True

    with patch(
        "salt.cloud.clouds.dimensiondata.show_instance",
        MagicMock(
            return_value={
                "state": True,
                "name": "foo",
                "public_ips": [],
                "private_ips": private_ips,
            }
        ),
    ):
        with patch(
            "salt.cloud.clouds.dimensiondata.preferred_ip",
            _preferred_ip(private_ips, [zero_ip]),
        ):
            with patch(
                "salt.cloud.clouds.dimensiondata.ssh_interface",
                MagicMock(return_value="private_ips"),
            ):
                assert dimensiondata._query_node_data(vm, data).public_ips == [zero_ip]



"""
Connection module for Amazon Data Pipeline

.. versionadded:: 2016.3.0

:depends: boto3
"""


import logging

import salt.utils.versions

log = logging.getLogger(__name__)

try:
    import boto3
    import botocore.exceptions

    boto3.set_stream_logger(level=logging.CRITICAL)
    logging.getLogger("botocore").setLevel(logging.CRITICAL)
    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False


def __virtual__():
    """
    Only load if boto3 libraries exists.
    """
    return salt.utils.versions.check_boto_reqs(check_boto=False)


def activate_pipeline(pipeline_id, region=None, key=None, keyid=None, profile=None):
    """
    Start processing pipeline tasks. This function is idempotent.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_datapipeline.activate_pipeline my_pipeline_id
    """
    client = _get_client(region, key, keyid, profile)
    r = {}
    try:
        client.activate_pipeline(pipelineId=pipeline_id)
        r["result"] = True
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        r["error"] = str(e)
    return r


def create_pipeline(
    name, unique_id, description="", region=None, key=None, keyid=None, profile=None
):
    """
    Create a new, empty pipeline. This function is idempotent.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_datapipeline.create_pipeline my_name my_unique_id
    """
    client = _get_client(region, key, keyid, profile)
    r = {}
    try:
        response = client.create_pipeline(
            name=name,
            uniqueId=unique_id,
            description=description,
        )
        r["result"] = response["pipelineId"]
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        r["error"] = str(e)
    return r


def delete_pipeline(pipeline_id, region=None, key=None, keyid=None, profile=None):
    """
    Delete a pipeline, its pipeline definition, and its run history. This function is idempotent.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_datapipeline.delete_pipeline my_pipeline_id
    """
    client = _get_client(region, key, keyid, profile)
    r = {}
    try:
        client.delete_pipeline(pipelineId=pipeline_id)
        r["result"] = True
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        r["error"] = str(e)
    return r


def describe_pipelines(pipeline_ids, region=None, key=None, keyid=None, profile=None):
    """
    Retrieve metadata about one or more pipelines.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_datapipeline.describe_pipelines ['my_pipeline_id']
    """
    client = _get_client(region, key, keyid, profile)
    r = {}
    try:
        r["result"] = client.describe_pipelines(pipelineIds=pipeline_ids)
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        r["error"] = str(e)
    return r


def get_pipeline_definition(
    pipeline_id, version="latest", region=None, key=None, keyid=None, profile=None
):
    """
    Get the definition of the specified pipeline.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_datapipeline.get_pipeline_definition my_pipeline_id
    """
    client = _get_client(region, key, keyid, profile)
    r = {}
    try:
        r["result"] = client.get_pipeline_definition(
            pipelineId=pipeline_id,
            version=version,
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        r["error"] = str(e)
    return r


def list_pipelines(region=None, key=None, keyid=None, profile=None):
    """
    Get a list of pipeline ids and names for all pipelines.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_datapipeline.list_pipelines profile=myprofile
    """
    client = _get_client(region, key, keyid, profile)
    r = {}
    try:
        paginator = client.get_paginator("list_pipelines")
        pipelines = []
        for page in paginator.paginate():
            pipelines += page["pipelineIdList"]
        r["result"] = pipelines
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        r["error"] = str(e)
    return r


def pipeline_id_from_name(name, region=None, key=None, keyid=None, profile=None):
    """
    Get the pipeline id, if it exists, for the given name.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_datapipeline.pipeline_id_from_name my_pipeline_name
    """
    r = {}
    result_pipelines = list_pipelines()
    if "error" in result_pipelines:
        return result_pipelines

    for pipeline in result_pipelines["result"]:
        if pipeline["name"] == name:
            r["result"] = pipeline["id"]
            return r
    r["error"] = "No pipeline found with name={}".format(name)
    return r


def put_pipeline_definition(
    pipeline_id,
    pipeline_objects,
    parameter_objects=None,
    parameter_values=None,
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Add tasks, schedules, and preconditions to the specified pipeline. This function is
    idempotent and will replace an existing definition.

    CLI Example:

    .. code-block:: bash

        salt myminion boto_datapipeline.put_pipeline_definition my_pipeline_id my_pipeline_objects
    """
    parameter_objects = parameter_objects or []
    parameter_values = parameter_values or []
    client = _get_client(region, key, keyid, profile)
    r = {}
    try:
        response = client.put_pipeline_definition(
            pipelineId=pipeline_id,
            pipelineObjects=pipeline_objects,
            parameterObjects=parameter_objects,
            parameterValues=parameter_values,
        )
        if response["errored"]:
            r["error"] = response["validationErrors"]
        else:
            r["result"] = response
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        r["error"] = str(e)
    return r


def _get_client(region, key, keyid, profile):
    """
    Get a boto connection to Data Pipeline.
    """
    session = _get_session(region, key, keyid, profile)
    if not session:
        log.error("Failed to get datapipeline client.")
        return None

    return session.client("datapipeline")


def _get_session(region, key, keyid, profile):
    """
    Get a boto3 session
    """
    if profile:
        if isinstance(profile, str):
            _profile = __salt__["config.option"](profile)
        elif isinstance(profile, dict):
            _profile = profile
        key = _profile.get("key", None)
        keyid = _profile.get("keyid", None)
        region = _profile.get("region", None)

    if not region and __salt__["config.option"]("datapipeline.region"):
        region = __salt__["config.option"]("datapipeline.region")

    if not region:
        region = "us-east-1"

    return boto3.session.Session(
        region_name=region,
        aws_secret_access_key=key,
        aws_access_key_id=keyid,
    )



"""
Tests for win_certutil state module
"""
import pytest

import salt.utils.files

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(scope="module")
def certutil(states):
    return states.certutil


@pytest.fixture(scope="module")
def certutil_mod(modules):
    return modules.certutil


@pytest.fixture(scope="module")
def cert_file(state_tree):
    # This is the binary contents of a self-signed cert for testing
    binary_data = (
        b"0\x82\x03\x0e0\x82\x01\xf6\xa0\x03\x02\x01\x02\x02\x10[\xe1\xcc]Q\xb7"
        b"\x8d\xbdI\xa0\xb7\xc0\rD\x80m0\r\x06\t*\x86H\x86\xf7\r\x01\x01\x0b"
        b"\x05\x000\x1a1\x180\x16\x06\x03U\x04\x03\x0c\x0fTestCertificate0\x1e"
        b"\x17\r220120174254Z\x17\r230120180254Z0\x1a1\x180\x16\x06\x03U\x04"
        b'\x03\x0c\x0fTestCertificate0\x82\x01"0\r\x06\t*\x86H\x86\xf7\r\x01'
        b"\x01\x01\x05\x00\x03\x82\x01\x0f\x000\x82\x01\n\x02\x82\x01\x01\x00"
        b"\xb8x@YBP\x9f\x9c\x0e\n\xad\xd0l6\xc4\x9c\x7f#\x97\xbck@b\\\xa1\x94"
        b"\xecR\x85Xq\xe4H\x0c\xfa\x1b]\xb8\x14\x14x\x05\xb7\xe6\xb6t\x07j\xda0"
        b"\xd0\xb5\xc8\xdf\xe8\xad\xeb4qa\x86\xefw\x19\xf0\x9a%\xb8!\x81\xc2"
        b"\xcbd\x81,\xbd\xe1a\x91\x822\nh\x88\x9d\xb7\x82 \xe8\x0f\x91\x13\xc8"
        b"\xc0xir\xf8\x90Yc\x8f3\xe9\xdc\xa3\xbc+\xea/\x02\n\x94\xde\xba\xbb"
        b"\xcb0\x98Z\xbc\xeeK\xab\xc5\xba,\x0f\x7f}6\xb9$|\xdd=\xdaN\xff]N\xe3"
        b"\xbd\x00\xee?H\xdav\xa9\x95\xb8Vd\xf9=\x01\x16K\xb8\xa0C%\x1e[\x18'"
        b"\xb4\x17Vi\xee\x97[\xf9\xa8MM\xfb\x88\x9fc\xbb\x08\xa7!\xc0U\xa8\xfc"
        b"\nx:\xbc\x8f\x14\x0eF\x1f\x85Ba\x8b\xa3\xd7\xc4<\xcaN\xd1;y\xd0\x1a"
        b"\xeb\xd2\x91c\x94\xee%\xc8\x82\x85\x92\x88\xec\x1d\nh\xa9q|E\x1a\xaf"
        b"\x16\x89!i\x19'\xb7t{\x11\xe8\xb8\xee\xa9\x97\xf4\x1c\xfa\x92-\x02"
        b"\x03\x01\x00\x01\xa3P0N0\x0e\x06\x03U\x1d\x0f\x01\x01\xff\x04\x04\x03"
        b"\x02\x05\xa00\x1d\x06\x03U\x1d%\x04\x160\x14\x06\x08+\x06\x01\x05\x05"
        b"\x07\x03\x02\x06\x08+\x06\x01\x05\x05\x07\x03\x010\x1d\x06\x03U\x1d"
        b"\x0e\x04\x16\x04\x14\xefy\x97r\x16\xadg\r\x85\xea\xfe\xa8y[29\x0b%"
        b"\xdfB0\r\x06\t*\x86H\x86\xf7\r\x01\x01\x0b\x05\x00\x03\x82\x01\x01"
        b"\x00\x93)\x0c$\xeb\xf7\x02\x9fSf^[\t2\xd3\xdf\xcc~b\xdd\xd3\x1e<\x91"
        b"\xbc\x93\x87Z\x8ciC/\x87\x85\xf4\x18\xe0j\xae\xf3\x1c\xa7\xab\xf7\xfd"
        b"\xd9\xeb\x11:}Ys\x8f\xc9\\\xea\x17\xbb\x957\x9b\xef\x17E]RwY\x10\x8b"
        b'\x08\xc5\xa6\xc9\x05[\xe7\x11\xf3"2\xd3\xca\xf6\x05\x8a2\xc1S\x1e\xf0'
        b"\xdb\xfa,\xfc\x80\xb88-!\x07\xe5\x81mc'\xca\x16@\x16\xf7\x9b\xc5"
        b"\x95V;$\x95\xeab\xea\x1eX\x1dU\x97\x87\xc0\x17\xd0n\x01c@\x88z\xec"
        b"\x9ep\x19\x02I\xf6\xe4\xddr\xc3(\xb9\x98\x97$\xb8\xf3g\x16\x05\xa7"
        b"\x04\xf7\x15\x9a\xed!\x02\xd76\xb2nC\x04}sV=,\xd5\x8e\xb8hG\x99\xcb-x"
        b"\x0e\x05h\xee;\xcdp\x13\xfc)\xdb\xa9o\xb0\x1c\x0e\x86\xb2\r\xc5.\xb1"
        b"\x036\t\xd3l&\xd1\x13\xc1\xc1\x12\xfb\xc0\xab<\xaf\x04\x0eIW\xb8<OD"
        b'\xfe"(U\xc2&\xa8\xd8\x9bkY\xdb~\xf8\xad\xb7\xa8Mu\xb6\xef\x89\xf2'
        b"\xbeM"
    )
    with pytest.helpers.temp_file(
        "TestCertificate.cer", directory=state_tree
    ) as cert_file:
        with salt.utils.files.fopen(str(cert_file), "wb") as fh:
            fh.write(binary_data)
        yield cert_file


@pytest.fixture(scope="module")
def invalid_cert_file(state_tree):
    with pytest.helpers.temp_file("Invalid.cer", directory=state_tree) as cert_file:
        with salt.utils.files.fopen(str(cert_file), "wb") as fh:
            fh.write(b"Invalid cert data")
        yield cert_file


@pytest.fixture(scope="function")
def clean_store(certutil_mod, cert_file):
    certutil_mod.del_store(source=str(cert_file), store="TrustedPublisher")
    serials = certutil_mod.get_stored_cert_serials(store="TrustedPublisher")
    assert "5be1cc5d51b78dbd49a0b7c00d44806d" not in serials
    yield
    certutil_mod.del_store(source=str(cert_file), store="TrustedPublisher")


@pytest.fixture(scope="function")
def populate_store(certutil_mod, cert_file):
    certutil_mod.add_store(source=str(cert_file), store="TrustedPublisher")
    serials = certutil_mod.get_stored_cert_serials(store="TrustedPublisher")
    assert "5be1cc5d51b78dbd49a0b7c00d44806d" in serials
    yield
    certutil_mod.del_store(source=str(cert_file), store="TrustedPublisher")


def test_add_store_non_existing_cert(certutil):
    """
    Test add_store when the certificate does not exist
    """
    ret = certutil.add_store(
        name="salt://non-existing.cer",
        store="TrustedPublisher",
    )
    assert ret.comment.startswith("Certificate file not found")
    assert ret.result is False


def test_add_store_invalid_cert(certutil, invalid_cert_file):
    """
    Test add_store with an invalid certificate
    """
    ret = certutil.add_store(name="salt://Invalid.cer", store="TrustedPublisher")
    assert ret.comment.startswith("Invalid certificate file")
    assert ret.result is False


def test_add_store_cert_already_present(certutil, cert_file, populate_store):
    """
    Test add_store when the certificate is already present
    """
    ret = certutil.add_store(
        name="salt://TestCertificate.cer",
        store="TrustedPublisher",
    )
    assert ret.comment.startswith("Certificate already present")
    assert ret.result is True


def test_add_store_cert_test_is_true(certutil, cert_file, clean_store):
    """
    Test add_store when test is True
    """
    ret = certutil.add_store(
        name="salt://TestCertificate.cer",
        store="TrustedPublisher",
        test=True,
    )
    assert ret.comment.startswith("Certificate will be added")
    assert ret.result is None


def test_add_store(certutil, cert_file, clean_store):
    """
    Test add_store
    """
    ret = certutil.add_store(
        name="salt://TestCertificate.cer",
        store="TrustedPublisher",
    )
    assert ret.comment.startswith("Added certificate")
    assert ret.result is True


def test_del_store_non_existing_cert(certutil):
    """
    Test del_store when the certificate does not exist
    """
    ret = certutil.del_store(
        name="salt://non-existing.cer",
        store="TrustedPublisher",
    )
    assert ret.comment.startswith("Certificate file not found")
    assert ret.result is False


def test_del_store_invalid_cert(certutil, invalid_cert_file):
    """
    Test del_store with an invalid certificate
    """
    ret = certutil.del_store(name="salt://Invalid.cer", store="TrustedPublisher")
    assert ret.comment.startswith("Invalid certificate file")
    assert ret.result is False


def test_del_store_cert_already_absent(certutil, cert_file, clean_store):
    """
    Test del_store when the certificate is already absent
    """
    ret = certutil.del_store(
        name="salt://TestCertificate.cer",
        store="TrustedPublisher",
    )
    assert ret.comment.startswith("Certificate already absent")
    assert ret.result is True


def test_del_store_cert_test_is_true(certutil, cert_file, populate_store):
    """
    Test del_store when test is True
    """
    ret = certutil.del_store(
        name="salt://TestCertificate.cer",
        store="TrustedPublisher",
        test=True,
    )
    assert ret.comment.startswith("Certificate will be removed")
    assert ret.result is None


def test_del_store(certutil, cert_file, populate_store):
    """
    Test del_store
    """
    ret = certutil.del_store(
        name="salt://TestCertificate.cer",
        store="TrustedPublisher",
    )
    assert ret.comment.startswith("Removed certificate")
    assert ret.result is True



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>

    Test cases for salt.states.network
"""

import logging

import pytest

import salt.states.network as network
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {network: {}}


class MockNetwork:
    """
    Mock network class
    """

    def __init__(self):
        pass

    @staticmethod
    def interfaces():
        """
        Mock interface method
        """
        ifaces = {
            "salt": {"up": 1},
            "lo": {"up": 1, "inet": [{"label": "lo"}, {"label": "lo:alias1"}]},
        }
        return ifaces


class MockGrains:
    """
    Mock Grains class
    """

    def __init__(self):
        pass

    @staticmethod
    def grains(lis, bol):
        """
        Mock grains method
        """
        return {"A": "B"}


def test_managed():
    """
    Test to ensure that the named interface is configured properly
    """
    with patch("salt.states.network.salt.utils.network", MockNetwork()), patch(
        "salt.states.network.salt.loader", MockGrains()
    ):
        ret = {"name": "salt", "changes": {}, "result": False, "comment": ""}

        change = {
            "interface": "--- \n+++ \n@@ -1 +1 @@\n-A\n+B",
            "status": "Interface salt restart to validate",
        }

        dunder_salt = {
            "ip.get_interface": MagicMock(
                side_effect=[AttributeError, "A", "A", "A", "A", "A"]
            ),
            "ip.build_interface": MagicMock(return_value="B"),
            "saltutil.refresh_grains": MagicMock(return_value=True),
        }

        with patch.dict(network.__salt__, dunder_salt):
            with patch.dict(
                network.__salt__,
                {"ip.get_bond": MagicMock(side_effect=AttributeError)},
            ):
                assert network.managed("salt", type="bond", test=True) == ret

            ret.update(
                {
                    "comment": (
                        "Interface salt is set to be"
                        " updated:\n--- \n+++ \n@@ -1 +1 @@\n-A\n+B"
                    ),
                    "result": None,
                }
            )
            assert network.managed("salt", type="stack", test=True) == ret

            ipupdown = MagicMock(return_value=True)
            with patch.dict(network.__salt__, {"ip.down": ipupdown, "ip.up": ipupdown}):
                ret.update(
                    {
                        "comment": "Interface salt updated.",
                        "result": True,
                        "changes": change,
                    }
                )
                assert network.managed("salt", type="stack") == ret

                with patch.dict(network.__grains__, {"A": True}):
                    ret.update(
                        {
                            "result": True,
                            "changes": {
                                "interface": "--- \n+++ \n@@ -1 +1 @@\n-A\n+B",
                                "status": "Interface salt down",
                            },
                        }
                    )
                    assert network.managed("salt", type="stack", enabled=False) == ret

                mock = MagicMock(return_value=True)
                with patch.dict(network.__salt__, {"ip.down": mock}):
                    with patch.dict(
                        network.__salt__, {"saltutil.refresh_modules": mock}
                    ):
                        change = {
                            "interface": "--- \n+++ \n@@ -1 +1 @@\n-A\n+B",
                            "status": "Interface lo:alias1 down",
                        }
                        ret.update(
                            {
                                "name": "lo:alias1",
                                "comment": "Interface lo:alias1 updated.",
                                "result": True,
                                "changes": change,
                            }
                        )
                        assert (
                            network.managed("lo:alias1", type="eth", enabled=False)
                            == ret
                        )


def test_routes():
    """
    Test to manage network interface static routes.
    """
    ret = {"name": "salt", "changes": {}, "result": False, "comment": ""}

    mock = MagicMock(side_effect=[AttributeError, False, False, "True", False, False])
    with patch.dict(network.__salt__, {"ip.get_routes": mock}):
        assert network.routes("salt") == ret

        mock = MagicMock(side_effect=[False, True, "", True, True])
        with patch.dict(network.__salt__, {"ip.build_routes": mock}):
            ret.update(
                {"result": True, "comment": "Interface salt routes are up to date."}
            )
            assert network.routes("salt", test="a") == ret

            ret.update(
                {
                    "comment": "Interface salt routes are set to be added.",
                    "result": None,
                }
            )
            assert network.routes("salt", test="a") == ret

            ret.update(
                {
                    "comment": (
                        "Interface salt routes are set to be"
                        " updated:\n--- \n+++ \n@@ -1,4 +0,0 @@\n-T\n-r"
                        "\n-u\n-e"
                    )
                }
            )
            assert network.routes("salt", test="a") == ret

            mock = MagicMock(side_effect=[AttributeError, True])
            with patch.dict(network.__salt__, {"ip.apply_network_settings": mock}):
                ret.update(
                    {
                        "changes": {"network_routes": "Added interface salt routes."},
                        "comment": "",
                        "result": False,
                    }
                )
                assert network.routes("salt") == ret

                ret.update(
                    {
                        "changes": {"network_routes": "Added interface salt routes."},
                        "comment": "Interface salt routes added.",
                        "result": True,
                    }
                )
                assert network.routes("salt") == ret


def test_system():
    """
    Test to ensure that global network settings
    are configured properly
    """
    ret = {"name": "salt", "changes": {}, "result": False, "comment": ""}

    with patch.dict(network.__opts__, {"test": True}):
        mock = MagicMock(side_effect=[AttributeError, False, False, "As"])
        with patch.dict(network.__salt__, {"ip.get_network_settings": mock}):
            assert network.system("salt") == ret

            mock = MagicMock(side_effect=[False, True, ""])
            with patch.dict(network.__salt__, {"ip.build_network_settings": mock}):
                ret.update(
                    {
                        "comment": "Global network settings are up to date.",
                        "result": True,
                    }
                )
                assert network.system("salt") == ret

                ret.update(
                    {
                        "comment": "Global network settings are set to be added.",
                        "result": None,
                    }
                )
                assert network.system("salt") == ret

                ret.update(
                    {
                        "comment": (
                            "Global network settings are set to"
                            " be updated:\n--- \n+++ \n@@ -1,2 +0,0"
                            " @@\n-A\n-s"
                        )
                    }
                )
                assert network.system("salt") == ret

    with patch.dict(network.__opts__, {"test": False}):
        mock = MagicMock(side_effect=[False, False])
        with patch.dict(network.__salt__, {"ip.get_network_settings": mock}):
            mock = MagicMock(side_effect=[True, True])
            with patch.dict(network.__salt__, {"ip.build_network_settings": mock}):
                mock = MagicMock(side_effect=[AttributeError, True])
                with patch.dict(network.__salt__, {"ip.apply_network_settings": mock}):
                    ret.update(
                        {
                            "changes": {
                                "network_settings": "Added global network settings."
                            },
                            "comment": "",
                            "result": False,
                        }
                    )
                    assert network.system("salt") == ret

                    ret.update(
                        {
                            "changes": {
                                "network_settings": "Added global network settings."
                            },
                            "comment": "Global network settings are up to date.",
                            "result": True,
                        }
                    )
                    assert network.system("salt") == ret
